#!/usr/bin/env python
from pg_chameleon import sql_token


#statement="""alter table test_pk drop primary key;"""
#statement="""ALTER TABLE test ADD COLUMN `dkdkd` timestamp NULL;"""
#statement="""create table test_pk (id int ,PRIMARY KEY  (id) ); """
#statement="""alter table test change   date_create_new date_create_new timestamp;"""
#statement = """alter table test add column `test_default` varchar(30) not null default 20  """
#statement = """ALTER TABLE test
#ADD COLUMN `count` SMALLINT(6) NULL ,
#ADD COLUMN `log` VARCHAR(12) NULL default 'blah' AFTER `count`,
#ADD COLUMN new_enum ENUM('asd','r') NULL AFTER `log`,
#ADD COLUMN status INT(10) UNSIGNED NULL AFTER `new_enum`
#"""

statement = """ALTER TABLE foo DROP FOREIGN KEY fk_trigger_bar,ADD COLUMN `count` SMALLINT(6) NULL;"""
statement = """ALTER TABLE test
ADD `count` SMALLINT(6) NULL ,
ADD COLUMN `log` VARCHAR(12) default 'blah' NULL AFTER `count`,
ADD COLUMN new_enum ENUM('asd','r') NULL AFTER `log`,
ADD COLUMN status INT(10) UNSIGNED NULL AFTER `new_enum`,
ADD COLUMN mydate datetime NULL AFTER `status`,
ADD COLUMN mytstamp timestamp NULL AFTER `status`,
DROP FOREIGN            KEY fk_trigger_bar,
add primary key,
drop unique index asdf
"""
statement="""
CREATE TABLE film_text (
  film_id SMALLINT NOT NULL,
  title VARCHAR(255) NOT NULL,
  description TEXT,
  PRIMARY KEY  (film_id),
   KEY idx_title_description (title,description)
)ENGINE=InnoDB DEFAULT CHARSET=utf8
"""

statement="""
CREATE TABLE film_text (
  film_id SMALLINT NOT NULL PRIMARY KEY
)ENGINE=InnoDB DEFAULT CHARSET=utf8
"""
#statement="""RENAME TABLE `sakila`.`test_partition` TO `sakila`.`_test_partition_old`, `_test_partition_new` TO `test_partition`;"""
#statement="""RENAME TABLE test_partition TO _test_partition_old, _test_partition_new TO test_partition; """
#statement="""RENAME TABLE sakila.test_partition TO sakila._test_partition_old, sakila._test_partition_new TO sakila.test_partition ;"""
#statement="""RENAME TABLE `sakila`.`test_partition` TO `sakila`.`_test_partition_old`, `sakila`.`_test_partition_new` TO `sakila`.`test_partition`;"""
#statement = """create table blah(id integer(30) not null auto_increment, datevalue datetime,primary key (id,datevalue))"""
#statement = """alter table dd add column(foo varchar(30)); alter table dd add column foo varchar(30);"""
#statement = """create table test_tiny(id int(4) auto_increment, value tinyint(1), unique key(id),unique key (id,value),unique key (value,id)); """


statement=""" 

-- drop table
DROP TABLE `test`;
-- create table
CREATE   TABLE `test` (
  store_id TINYINT UNSIGNED NULL AUTO_INCREMENT,
  manager_staff_id TINYINT UNSIGNED NOT NULL,
  address_id SMALLINT UNSIGNED NOT NULL,
  `address_txt` varchar (30) NOT NULL default 'default_t;ext',
  `address_dp` double precision (30,2) NOT NULL,
  `test_enum` enum ('a','b'),
  size ENUM('x-small', 'small', 'medium', 'large', 'x-large'),
  last_update TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP,
  PRIMARY KEY  (store_id,address_id),
  UNIQUE KEY idx_unique_manager (manager_staff_id),
  KEY idx_fk_address_id2 (address_id),
  index
  idx_fk_address_id (address_id,store_id)
)ENGINE=InnoDB DEFAULT CHARSET=utf8;


ALTER TABLE test
ADD COLUMN `count` SMALLINT(6) NOT NULL default 3 AFTER `test_enum`,
ADD COLUMN `log` VARCHAR(12) NOT NULL AFTER `count`,
ADD COLUMN new_enum ENUM('asd','r') NOT NULL AFTER `log`,
ADD COLUMN status INT(10) UNSIGNED NOT NULL AFTER `new_enum`;


ALTER TABLE `test`
DROP COLUMN `count` ,
ADD COLUMN newstatus INT(10) UNSIGNED NOT NULL AFTER `log`;

ALTER TABLE `test` DROP PRIMARY KEY;

				"""
statement="""ALTER TABLE t_user_info ADD (
group_id INT(11) UNSIGNED DEFAULT NULL,
contact_phone VARCHAR(20) DEFAULT NULL
);"""
statement = """ALTER TABLE foo RENAME TO bar;"""
statement = """RENAME TABLE `sakila`.`test_partition` TO `sakila`.`_test_partition_old`, `_test_partition_new` TO `test_partition`;"""
#statement="""ALTER TABLE foo MODIFY bar INT UNSIGNED DEFAULT NULL;"""
#statement="""ALTER TABLE foo change bar bar INT UNSIGNED;"""
statement="""ALTER TABLE `some_sch`.`my_great_table` CHANGE COLUMN `IMEI` `IMEI` VARCHAR(255) NULL DEFAULT NULL COMMENT 'IMEI datatype changed'"""
token_sql=sql_token()
token_sql.parse_sql(statement)
print (token_sql.tokenised)
#for token in token_sql.tokenised:
	#print (token)
#	for column in token["columns"]:
#		print(column)
	#else:	




#!/usr/bin/env python

import unittest

from pg_chameleon.lib.sql_util import sql_token


tokenizer = sql_token()


class TestTokenizer(unittest.TestCase):
    param_list = [
        (
            "column_definition",
            "id int PRIMARY KEY",
            dict(column_name="id", data_type="int",
                 dimensions=None, enum_list=None, extras=["PRIMARY KEY"])
        ),
        (
            "any_key_definition",
            "CONSTRAINT pk_id PRIMARY KEY (id)",
            dict(index_name="PRIMARY", index_columns=["id"], non_unique=0)
        ),
        (
            "any_key_definition",
            "CONSTRAINT pk_id1_id2 PRIMARY KEY (id1,id2)",
            dict(index_name="PRIMARY", index_columns=["id1", "id2"], non_unique=0)
        ),
        (
            "any_key_definition",
            "PRIMARY\n KEY(id1, id2, `id3`, `id 4` )",
            dict(index_name="PRIMARY", index_columns=["id1", "id2", "id3", "id 4"], non_unique=0)
        ),
        (
            "any_key_definition",
            "UNIQUE KEY (id1, id2)",
            dict(index_name="UNIQUE", index_columns=["id1", "id2"], non_unique=0)
        ),
        (
            "any_key_definition",
            "UNIQUE KEY (id1)",
            dict(index_name="UNIQUE", index_columns=["id1"], non_unique=0)
        ),
        (
            "any_key_definition",
            "UNIQUE (id1, id2)",
            dict(index_name="UNIQUE", index_columns=["id1", "id2"], non_unique=0)
        ),
        (
            "any_key_definition",
            "KEY (id1)",
            dict(index_name="INDEX", index_columns=["id1"], non_unique=1)
        ),
        (
            "any_key_definition",
            "FULLTEXT KEY asdf (id)",
            dict(index_name="OTHER", index_columns=["id"], non_unique=1),
        ),
        (
            "any_key_definition",
            "FOREIGN KEY (column1) REFERENCES table2 (column2)",
            dict(index_name="FOREIGN", index_columns=["column1"], non_unique=1)
        ),
        (
            "column_definition",
            """film_rating smallint not null
            DEFAULT 3""",
            dict(column_name="film_rating", data_type="smallint",
                 dimensions=None, enum_list=None,
                 extras=["NOT NULL", ["DEFAULT", "3"]])
        ),
        (
            "column_definition",
            """film_rating
            ENUM ('1 star', '2 star', '3 star', '4 star', '5 star')
            DEFAULT '3 star'""",
            dict(column_name="film_rating", data_type="enum",
                 dimensions=None, enum_list=["1 star", "2 star", "3 star", "4 star", "5 star"],
                 extras=[["DEFAULT", "'3 star'"]])
        )
    ]

    def test_parses_successfully(self):
        for parser_name, string, expected in self.param_list:
            parser = getattr(tokenizer, parser_name)
            with self.subTest(
                msg=f"Checking parser {parser_name}", string=string, expected=expected
            ):
                output = parser.parse(string)
                self.assertEqual(output, expected)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
from pkg_resources import get_distribution
__version__ = get_distribution('pg_chameleon')
import argparse
from pg_chameleon import replica_engine

commands = [
    'show_config',
    'show_sources',
    'show_status',
    'create_replica_schema',
    'drop_replica_schema',
    'upgrade_replica_schema',
    'add_source',
    'drop_source',
    'init_replica',
    'enable_replica',
    'update_schema_mappings',
    'refresh_schema',
    'sync_tables',
    'start_replica',
    'stop_replica',
    'detach_replica',
    'set_configuration_files',
    'show_errors',
    'run_maintenance',
    'stop_all_replicas'
    ]

command_help = ','.join(commands)
config_help = """Specifies the configuration to use without the suffix yml. If  the parameter is omitted then ~/.pg_chameleon/configuration/default.yml is used"""
schema_help = """Specifies the schema within a source. If omitted all schemas for the given source are affected by the command. Requires the argument --source to be specified"""
source_help = """Specifies the source within a configuration. If omitted all sources are affected by the command."""
tables_help = """Specifies the tables within a source . If omitted all tables are affected by the command."""
logid_help = """Specifies the log id entry for displaying the error details"""
debug_help = """Forces the debug mode with logging on stdout and log level debug."""
version_help = """Displays pg_chameleon's installed  version."""
rollbar_help = """Overrides the level for messages to be sent to rolllbar. One of: "critical", "error", "warning", "info". The Default is "info" """
full_help = """When specified with run_maintenance the switch performs a vacuum full instead of a normal vacuum. """
truncate_help = """Truncate the existing tables instead of replacing them."""

parser = argparse.ArgumentParser(description='Command line for pg_chameleon.',  add_help=True)
parser.add_argument('command', type=str, help=command_help)
parser.add_argument('--config', type=str,  default='default',  required=False, help=config_help)
parser.add_argument('--schema', type=str,  default='*',  required=False, help=schema_help)
parser.add_argument('--source', type=str,  default='*',  required=False, help=source_help)
parser.add_argument('--tables', type=str,  default='*',  required=False, help=tables_help)
parser.add_argument('--logid', type=str,  default='*',  required=False, help=logid_help)
parser.add_argument('--debug', default=False, required=False, help=debug_help, action='store_true')
parser.add_argument('--version', action='version', help=version_help,version='{version}'.format(version=__version__))
parser.add_argument('--rollbar-level', type=str, default="info", required=False, help=rollbar_help)
parser.add_argument('--full', default=False, required=False, help=full_help, action='store_true')
args = parser.parse_args()


replica = replica_engine(args)
if args.debug:
    getattr(replica, args.command)()
else:
    try:
        getattr(replica, args.command)()
    except AttributeError:
        print("ERROR - Invalid command" )
        print(command_help)



import time
import sys
import io
import pymysql
import codecs
import binascii
from pymysqlreplication import BinLogStreamReader
from pymysqlreplication.event import QueryEvent, GtidEvent, HeartbeatLogEvent
from pymysqlreplication.row_event import DeleteRowsEvent,UpdateRowsEvent,WriteRowsEvent
from pymysqlreplication.event import RotateEvent
from pg_chameleon import sql_token
from os import remove
import re
class mysql_source(object):
    def __init__(self):
        """
            Class constructor, the method sets the class variables and configure the
            operating parameters from the args provided t the class.
        """
        self.statement_skip = ['BEGIN', 'COMMIT']
        self.schema_tables = {}
        self.schema_mappings = {}
        self.schema_loading = {}
        self.schema_list = []
        self.hexify_always = ['blob', 'tinyblob', 'mediumblob','longblob','binary','varbinary']
        self.spatial_datatypes = ['point','geometry','linestring','polygon', 'multipoint', 'multilinestring', 'geometrycollection']
        self.schema_only = {}
        self.gtid_mode = False
        self.gtid_enable = False
        self.copy_table_data = True


    def __del__(self):
        """
            Class destructor, tries to disconnect the mysql connection.
        """
        self.disconnect_db_unbuffered()
        self.disconnect_db_buffered()

    def __check_mysql_config(self):
        """
            The method check if the mysql configuration is compatible with the replica requirements.
            If all the configuration requirements are met then the return value is True.
            Otherwise is false.
            The parameters checked are:
            log_bin - ON if the binary log is enabled if we are on a vanilla mysql
            binlog_format - must be ROW , otherwise the replica won't get the data
            binlog_row_image - must be FULL, otherwise the row image will be incomplete

            The method checks the function AURORA_VERSION() and if the query doesn't error then skips the log_bin parameter check.
        """
        sql_aurora = """select AURORA_VERSION() ver;"""
        try:
            self.cursor_buffered.execute(sql_aurora)
            skip_log_bin_check = True
        except:
            skip_log_bin_check = False

        if self.gtid_enable:
            sql_log_bin = """SHOW GLOBAL VARIABLES LIKE 'gtid_mode';"""
            self.cursor_buffered.execute(sql_log_bin)
            variable_check = self.cursor_buffered.fetchone()
            if variable_check:
                gtid_mode = variable_check["Value"]
                if gtid_mode.upper() == 'ON':
                    self.gtid_mode = True
                    sql_uuid = """SHOW SLAVE STATUS;"""
                    self.cursor_buffered.execute(sql_uuid)
                    slave_status = self.cursor_buffered.fetchall()
                    if len(slave_status)>0:
                        gtid_set=slave_status[0]["Retrieved_Gtid_Set"]
                    else:
                        sql_uuid = """SHOW GLOBAL VARIABLES LIKE 'server_uuid';"""
                        self.cursor_buffered.execute(sql_uuid)
                        server_uuid = self.cursor_buffered.fetchone()
                        gtid_set = server_uuid["Value"]
                    self.gtid_uuid = gtid_set.split(':')[0]

            else:
                self.gtid_mode = False
        else:
            self.gtid_mode = False

        sql_log_bin = """SHOW GLOBAL VARIABLES LIKE 'log_bin';"""
        self.cursor_buffered.execute(sql_log_bin)
        variable_check = self.cursor_buffered.fetchone()
        log_bin = variable_check["Value"]

        sql_log_bin = """SHOW GLOBAL VARIABLES LIKE 'binlog_format';"""
        self.cursor_buffered.execute(sql_log_bin)
        variable_check = self.cursor_buffered.fetchone()
        binlog_format = variable_check["Value"]

        sql_log_bin = """SHOW GLOBAL VARIABLES LIKE 'binlog_row_image';"""
        self.cursor_buffered.execute(sql_log_bin)
        variable_check = self.cursor_buffered.fetchone()
        if variable_check:
            binlog_row_image = variable_check["Value"]
        else:
            binlog_row_image = 'FULL'

        if (log_bin.upper() == 'ON' or skip_log_bin_check) and binlog_format.upper() == 'ROW' and binlog_row_image.upper() == 'FULL':
            self.replica_possible = True
        else:
            self.replica_possible = False
            self.pg_engine.set_source_status("error")
            self.logger.error("The MySQL configuration does not allow the replica. Exiting now")
            self.logger.error("Source settings - log_bin %s, binlog_format %s, binlog_row_image %s" % (log_bin.upper(),  binlog_format.upper(), binlog_row_image.upper() ))
            self.logger.error("Mandatory settings - log_bin ON, binlog_format ROW, binlog_row_image FULL (only for MySQL 5.6+) ")
            sys.exit()


    def connect_db_buffered(self):
        """
            The method creates a new connection to the mysql database.
            The connection is made using the dictionary type cursor factory, which is buffered.
        """
        db_conn = self.source_config["db_conn"]
        db_conn = {key:str(value) for key, value in db_conn.items()}
        db_conn["port"] = int(db_conn["port"])
        db_conn["connect_timeout"] = int(db_conn["connect_timeout"])

        self.conn_buffered=pymysql.connect(
            host = db_conn["host"],
            user = db_conn["user"],
            port = db_conn["port"],
            password = db_conn["password"],
            charset = db_conn["charset"],
            connect_timeout = db_conn["connect_timeout"],
            cursorclass=pymysql.cursors.DictCursor
        )
        self.charset = db_conn["charset"]
        self.cursor_buffered = self.conn_buffered.cursor()
        self.cursor_buffered_fallback = self.conn_buffered.cursor()

    def disconnect_db_buffered(self):
        """
            The method disconnects any connection  with dictionary type cursor from the mysql database.

        """
        try:
            self.conn_buffered.close()
        except:
            pass

    def connect_db_unbuffered(self):
        """
            The method creates a new connection to the mysql database.
            The connection is made using the unbuffered cursor factory.
        """
        db_conn = self.source_config["db_conn"]
        db_conn = {key:str(value) for key, value in db_conn.items()}
        db_conn["port"] = int(db_conn["port"])
        db_conn["connect_timeout"] = int(db_conn["connect_timeout"])
        self.conn_unbuffered=pymysql.connect(
            host = db_conn["host"],
            user = db_conn["user"],
            port = db_conn["port"],
            password = db_conn["password"],
            charset = db_conn["charset"],
            connect_timeout = db_conn["connect_timeout"],
            cursorclass=pymysql.cursors.SSCursor
        )
        self.charset = db_conn["charset"]
        self.cursor_unbuffered = self.conn_unbuffered.cursor()


    def disconnect_db_unbuffered(self):
        """
            The method disconnects any unbuffered connection from the mysql database.
        """
        try:
            self.conn_unbuffered.close()
        except:
            pass

    def __build_skip_events(self):
        """
            The method builds a class attribute self.skip_events. The attribute is a dictionary with the tables and schemas listed under the three kind of skippable events  (insert,delete,update) using
            the configuration parameter skip_events.
        """
        self.skip_events = None
        if "skip_events" in  self.source_config:
            skip_events  = self.source_config["skip_events"]
            self.skip_events = {}
            if "insert" in skip_events:
                self.skip_events["insert"] = skip_events["insert"]
            else:
                self.skip_events["insert"] = []

            if "update" in skip_events:
                self.skip_events["update"] = skip_events["update"]
            else:
                self.skip_events["update"] = []

            if "delete" in skip_events:
                self.skip_events["delete"] = skip_events["delete"]
            else:
                self.skip_events["delete"] = []



    def __build_table_exceptions(self):
        """
            The method builds two dictionaries from the limit_tables and skip tables values set for the source.
            The dictionaries are intended to be used in the get_table_list to cleanup the list of tables per schema.
            The method manages the particular case of when the class variable self.tables is set.
            In that case only the specified tables in self.tables will be synced. Should limit_tables be already
            set, then the resulting list is the intersection of self.tables and limit_tables.
        """
        self.limit_tables = {}
        self.skip_tables = {}
        limit_tables = self.source_config["limit_tables"]
        skip_tables = self.source_config["skip_tables"]

        if self.tables !='*':
            tables = [table.strip() for table in self.tables.split(',')]
            if limit_tables:
                limit_schemas = [table.split('.')[0] for table in limit_tables]
                limit_tables = [table for table in tables if table in limit_tables or table.split('.')[0] not in limit_schemas]
            else:
                limit_tables = tables
            self.schema_only = {table.split('.')[0] for table in limit_tables}


        if limit_tables:
            table_limit = [table.split('.') for table in limit_tables]
            for table_list in table_limit:
                list_exclude = []
                try:
                    list_exclude = self.limit_tables[table_list[0]]
                    list_exclude.append(table_list[1])
                except KeyError:
                    try:
                        list_exclude.append(table_list[1])
                    except IndexError:
                        pass

                self.limit_tables[table_list[0]]  = list_exclude
        if skip_tables:
            table_skip = [table.split('.') for table in skip_tables]
            for table_list in table_skip:
                list_exclude = []
                try:
                    list_exclude = self.skip_tables[table_list[0]]
                    list_exclude.append(table_list[1])
                except KeyError:
                    try:
                        list_exclude.append(table_list[1])
                    except:
                        pass
                self.skip_tables[table_list[0]]  = list_exclude


    def get_table_list(self):
        """
            The method pulls the table list from the information_schema.
            The list is stored in a dictionary  which key is the table's schema.
        """
        sql_tables="""
            SELECT
                table_name as table_name
            FROM
                information_schema.TABLES
            WHERE
                    table_type='BASE TABLE'
                AND table_schema=%s
            ;
        """
        for schema in self.schema_list:
            self.cursor_buffered.execute(sql_tables, (schema))
            table_list = [table["table_name"] for table in self.cursor_buffered.fetchall()]
            try:
                limit_tables = self.limit_tables[schema]
                if len(limit_tables) > 0:
                    table_list = [table for table in table_list if table in limit_tables]
            except KeyError:
                pass
            try:
                skip_tables = self.skip_tables[schema]
                if len(skip_tables) > 0:
                    table_list = [table for table in table_list if table not in skip_tables]
            except KeyError:
                pass

            self.schema_tables[schema] = table_list

    def create_destination_schemas(self):
        """
            Creates the loading schemas in the destination database and associated tables listed in the dictionary
            self.schema_tables.
            The method builds a dictionary which associates the destination schema to the loading schema.
            The loading_schema is named after the destination schema plus with the prefix _ and the _tmp suffix.
            As postgresql allows, by default up to 64  characters for an identifier, the original schema is truncated to 59 characters,
            in order to fit the maximum identifier's length.
            The mappings are stored in the class dictionary schema_loading.
            If the source parameter keep_existing_schema is set to true the method doesn't create the schemas.
            Instead assumes the schema and the tables are already there.
        """
        if self.keep_existing_schema:
            self.logger.debug("Keep existing schema is set to True. Skipping the schema creation." )
            for schema in self.schema_list:
                destination_schema = self.schema_mappings[schema]
                self.schema_loading[schema] = {'destination':destination_schema, 'loading':destination_schema}
        else:
            for schema in self.schema_list:
                destination_schema = self.schema_mappings[schema]
                loading_schema = "_%s_tmp" % destination_schema[0:59]
                self.schema_loading[schema] = {'destination':destination_schema, 'loading':loading_schema}
                self.logger.debug("Creating the loading schema %s." % loading_schema)
                self.pg_engine.create_database_schema(loading_schema)
                self.logger.debug("Creating the destination schema %s." % destination_schema)
                self.pg_engine.create_database_schema(destination_schema)

    def drop_loading_schemas(self):
        """
            The method drops the loading schemas from the destination database.
            The drop is performed on the schemas generated in create_destination_schemas.
            The method assumes the class dictionary schema_loading is correctly set.
        """
        for schema in self.schema_loading:
            loading_schema = self.schema_loading[schema]["loading"]
            self.logger.debug("Dropping the schema %s." % loading_schema)
            self.pg_engine.drop_database_schema(loading_schema, True)

    def get_table_metadata(self, table, schema):
        """
            The method builds the table's metadata querying the information_schema.
            The data is returned as a dictionary.

            :param table: The table name
            :param schema: The table's schema
            :return: table's metadata as a cursor dictionary
            :rtype: dictionary
        """
        sql_metadata="""
            SELECT
                column_name as column_name,
                column_default as column_default,
                ordinal_position as ordinal_position,
                data_type as data_type,
                column_type as column_type,
                character_maximum_length as character_maximum_length,
                extra as extra,
                column_key as column_key,
                is_nullable as is_nullable,
                numeric_precision as numeric_precision,
                numeric_scale as numeric_scale,
                CASE
                    WHEN data_type="enum"
                THEN
                    SUBSTRING(COLUMN_TYPE,5)
                END AS enum_list
            FROM
                information_schema.COLUMNS
            WHERE
                    table_schema=%s
                AND	table_name=%s
            ORDER BY
                ordinal_position
            ;
        """
        self.cursor_buffered.execute(sql_metadata, (schema, table))
        table_metadata=self.cursor_buffered.fetchall()
        return table_metadata

    def get_foreign_keys_metadata(self):
        """
            The method collects the foreign key metadata for the detach replica process.
            Currently doesn't get the ON UPDATE/ON DELETE triggers
        """
        self.__init_sync()
        schema_replica = "'%s'"  % "','".join([schema.strip() for schema in self.sources[self.source]["schema_mappings"]])
        self.logger.info("retrieving foreign keys metadata for schemas %s" % schema_replica)
        sql_fkeys = """
            SELECT
                kc.table_name as table_name,
                kc.table_schema as table_schema,
                CASE WHEN (
                                SELECT
                                    count(1)
                                FROM
                                    information_schema.referential_constraints r
                                WHERE
                                        rc.constraint_name=r.constraint_name
                                    AND  rc.constraint_schema=r.constraint_schema
                            )>1
                THEN
                    concat(substring(kc.constraint_name,1,59),'_',SUBSTRING(md5(uuid()),1,4))
                ELSE
                    kc.constraint_name
                END as constraint_name,
                kc.referenced_table_name as referenced_table_name,
                kc.referenced_table_schema as referenced_table_schema,
                group_concat(DISTINCT concat('"',kc.column_name,'"') ORDER BY POSITION_IN_UNIQUE_CONSTRAINT) as fk_cols,
                group_concat(DISTINCT concat('"',kc.referenced_column_name,'"') ORDER BY POSITION_IN_UNIQUE_CONSTRAINT) as ref_columns,
                concat('ON DELETE ',rc.delete_rule) AS on_delete,
                concat('ON UPDATE ',rc.update_rule) AS on_update
            FROM
                information_schema.key_column_usage kc
                INNER JOIN information_schema.referential_constraints rc
                ON
                        rc.table_name=kc.table_name
                    AND rc.constraint_schema=kc.table_schema
                    AND rc.constraint_name=kc.constraint_name
            WHERE
                    kc.table_schema in (%s)
                AND kc.referenced_table_name IS NOT NULL
                AND kc.referenced_table_schema in (%s)
            GROUP BY
                kc.table_name,
                kc.constraint_name,
                kc.referenced_table_name,
                kc.table_schema,
                kc.referenced_table_schema,
                rc.delete_rule,
                rc.update_rule,
                rc.constraint_name,
                rc.constraint_schema
            ORDER BY
                kc.table_name
            ;

        """ % (schema_replica, schema_replica)
        self.cursor_buffered.execute(sql_fkeys)
        fkey_list=self.cursor_buffered.fetchall()
        self.disconnect_db_buffered()
        return fkey_list

    def create_destination_tables(self):
        """
            The method creates the destination tables in the loading schema.
            The tables names are looped using the values stored in the class dictionary schema_tables.
        """
        for schema in self.schema_tables:
            table_list = self.schema_tables[schema]
            for table in table_list:
                table_metadata = self.get_table_metadata(table, schema)
                self.pg_engine.create_table(table_metadata, table, schema, 'mysql')


    def generate_select_statements(self, schema, table):
        """
            The generates the csv output and the statements output for the given schema and table.
            The method assumes there is a buffered database connection active.

            :param schema: the origin's schema
            :param table: the table name
            :return: the select list statements for the copy to csv and  the fallback to inserts.
            :rtype: dictionary
        """
        select_columns = {}
        sql_select="""
            SELECT
                CASE
                    WHEN
                        data_type IN ('"""+"','".join(self.hexify)+"""')
                    THEN
                        concat('hex(',column_name,')')
                    WHEN
                        data_type IN ('bit')
                    THEN
                        concat('cast(`',column_name,'` AS unsigned)')
                    WHEN
                        data_type IN ('datetime','timestamp','date')
                    THEN
                        concat('nullif(`',column_name,'`,cast("0000-00-00 00:00:00" as date))')
                    WHEN
                        data_type IN ('"""+"','".join(self.spatial_datatypes)+"""')
                    THEN
                        concat('ST_AsText(',column_name,')')

                ELSE
                    concat('cast(`',column_name,'` AS char CHARACTER SET """+ self.charset +""")')
                END
                AS select_csv,
                CASE
                    WHEN
                        data_type IN ('"""+"','".join(self.hexify)+"""')
                    THEN
                        concat('hex(',column_name,') AS','`',column_name,'`')
                    WHEN
                        data_type IN ('bit')
                    THEN
                        concat('cast(`',column_name,'` AS unsigned) AS','`',column_name,'`')
                    WHEN
                        data_type IN ('datetime','timestamp','date')
                    THEN
                        concat('nullif(`',column_name,'`,cast("0000-00-00 00:00:00" as date)) AS `',column_name,'`')
                    WHEN
                        data_type IN ('"""+"','".join(self.spatial_datatypes)+"""')
                    THEN
                        concat('ST_AsText(',column_name,') AS','`',column_name,'`')

                ELSE
                    concat('cast(`',column_name,'` AS char CHARACTER SET """+ self.charset +""") AS','`',column_name,'`')

                END
                AS select_stat,
                column_name as column_name
            FROM
                information_schema.COLUMNS
            WHERE
                table_schema=%s
                AND 	table_name=%s
            ORDER BY
                ordinal_position
            ;
        """
        self.cursor_buffered.execute(sql_select, (schema, table))
        select_data = self.cursor_buffered.fetchall()
        select_csv = ["COALESCE(REPLACE(%s, '\"', '\"\"'),'NULL') " % statement["select_csv"] for statement in select_data]
        select_stat = [statement["select_csv"] for statement in select_data]
        column_list = ['"%s"' % statement["column_name"] for statement in select_data]
        select_columns["select_csv"] = "REPLACE(CONCAT('\"',CONCAT_WS('\",\"',%s),'\"'),'\"NULL\"','NULL')" % ','.join(select_csv)
        select_columns["select_stat"]  = ','.join(select_stat)
        select_columns["column_list"]  = ','.join(column_list)
        return select_columns


    def begin_tx(self):
        """
            The method sets the isolation level to repeatable read and begins a transaction
        """
        self.logger.debug("set isolation level")
        self.cursor_unbuffered.execute("SET TRANSACTION ISOLATION LEVEL REPEATABLE READ")
        self.logger.debug("beginning transaction")
        self.cursor_unbuffered.execute("BEGIN")

    def end_tx(self):
        """
            The method ends the current transaction by rollback
            - We should never have changed source anyway
        """
        self.logger.debug("rolling back")
        self.cursor_unbuffered.execute("ROLLBACK")

    def make_tx_snapshot(self, schema, table):
        """
            The method forces creation of transaction snapshot by making a read of
            one row from the source table and discarding it
        """
        self.logger.debug("reading and discarding 1 row from `%s`.`%s`" % (schema, table))
        self.cursor_unbuffered.execute("SELECT * FROM `%s`.`%s` LIMIT 1" % (schema, table))

    def lock_table(self, schema, table):
        """
            The method flushes the given table with read lock.
            The method assumes there is a database connection active.

            :param schema: the origin's schema
            :param table: the table name

        """
        self.logger.debug("locking the table `%s`.`%s`" % (schema, table) )
        sql_lock = "FLUSH TABLES `%s`.`%s` WITH READ LOCK;" %(schema, table)
        self.logger.debug("collecting the master's coordinates for table `%s`.`%s`" % (schema, table) )
        self.cursor_buffered.execute(sql_lock)

    def unlock_tables(self):
        """
            The method unlocks all tables
        """
        self.logger.debug("unlocking the tables")
        sql_unlock = "UNLOCK TABLES;"
        self.cursor_buffered.execute(sql_unlock)

    def get_master_coordinates(self):
        """
            The method gets the master's coordinates and return them stored in a dictionary.
            The method assumes there is a database connection active.

            :return: the master's log coordinates for the given table
            :rtype: dictionary
        """
        if not self.conn_buffered.open:
            self.connect_db_buffered()
        sql_master = "SHOW MASTER STATUS;"
        self.cursor_buffered.execute(sql_master)
        master_status = self.cursor_buffered.fetchall()
        return master_status

    def copy_data(self, schema, table):
        """
            The method copy the data between the origin and destination table.
            The method locks the table read only mode and  gets the log coordinates which are returned to the calling method.

            :param schema: the origin's schema
            :param table: the table name
            :return: the log coordinates for the given table
            :rtype: dictionary
        """
        slice_insert = []
        loading_schema = self.schema_loading[schema]["loading"]
        self.connect_db_buffered()

        self.logger.debug("estimating rows in %s.%s" % (schema , table))
        sql_rows = """
            SELECT
                table_rows as table_rows,
                CASE
                    WHEN avg_row_length>0
                    then
                        round(({}/avg_row_length))
                ELSE
                    0
                END as copy_limit,
                transactions
            FROM
                information_schema.TABLES,
                information_schema.ENGINES
            WHERE
                    table_schema=%s
                AND	table_type='BASE TABLE'
                AND table_name=%s
                AND TABLES.engine = ENGINES.engine
            ;
        """
        sql_rows = sql_rows.format(self.copy_max_memory)
        self.cursor_buffered.execute(sql_rows, (schema, table))
        count_rows = self.cursor_buffered.fetchone()
        total_rows = count_rows["table_rows"]
        copy_limit = int(count_rows["copy_limit"])
        table_txs = count_rows["transactions"] == "YES"
        if copy_limit == 0:
            copy_limit = 1000000
        num_slices = int(total_rows//copy_limit)
        range_slices = list(range(num_slices+1))
        total_slices = len(range_slices)
        slice = range_slices[0]
        self.logger.debug("The table %s.%s will be copied in %s  estimated slice(s) of %s rows, using a transaction %s"  % (schema, table, total_slices, copy_limit, table_txs))
        out_file = '%s/%s_%s.csv' % (self.out_dir, schema, table )
        self.lock_table(schema, table)
        master_status = self.get_master_coordinates()

        select_columns = self.generate_select_statements(schema, table)
        csv_data = ""
        sql_csv = "SELECT %s as data FROM `%s`.`%s`;" % (select_columns["select_csv"], schema, table)
        column_list = select_columns["column_list"]
        self.logger.debug("Executing query for table %s.%s"  % (schema, table ))
        self.connect_db_unbuffered()
        if table_txs:
            self.begin_tx()
        self.cursor_unbuffered.execute(sql_csv)
        if table_txs:
            self.unlock_tables()
        while True:
            csv_results = self.cursor_unbuffered.fetchmany(copy_limit)
            if len(csv_results) == 0:
                break
            csv_data="\n".join(d[0] for d in csv_results )

            if self.copy_mode == 'direct':
                csv_file = io.StringIO()
                csv_file.write(csv_data)
                csv_file.seek(0)

            if self.copy_mode == 'file':
                csv_file = codecs.open(out_file, 'wb', self.charset)
                csv_file.write(csv_data)
                csv_file.close()
                csv_file = open(out_file, 'rb')
            try:
                self.pg_engine.copy_data(csv_file, loading_schema, table, column_list)
            except:
                self.logger.info("Table %s.%s error in PostgreSQL copy, saving slice number for the fallback to insert statements " %  (loading_schema, table ))
                slice_insert.append(slice)

            self.print_progress(slice+1,total_slices, schema, table)
            slice+=1

            csv_file.close()
        if len(slice_insert)>0:
            ins_arg={}
            ins_arg["slice_insert"] = slice_insert
            ins_arg["table"] = table
            ins_arg["schema"] = schema
            ins_arg["select_stat"] = select_columns["select_stat"]
            ins_arg["column_list"] = column_list
            ins_arg["copy_limit"] = copy_limit
            self.insert_table_data(ins_arg)


        if table_txs:
            self.end_tx()
        else:
            self.unlock_tables()
        self.cursor_unbuffered.close()
        self.disconnect_db_unbuffered()
        self.disconnect_db_buffered()

        try:
            remove(out_file)
        except:
            pass
        return master_status

    def insert_table_data(self, ins_arg):
        """
            This method is a fallback procedure whether copy_table_data fails.
            The ins_args is a list with the informations required to run the select for building the insert
            statements and the slices's start and stop.
            The process is performed in memory and can take a very long time to complete.

            :param pg_engine: the postgresql engine
            :param ins_arg: the list with the insert arguments (slice_insert, schema, table, select_stat,column_list, copy_limit)
        """
        slice_insert= ins_arg["slice_insert"]
        table = ins_arg["table"]
        schema = ins_arg["schema"]
        select_stat = ins_arg["select_stat"]
        column_list = ins_arg["column_list"]
        copy_limit = ins_arg["copy_limit"]
        self.connect_db_unbuffered()
        loading_schema = self.schema_loading[schema]["loading"]
        num_insert = 1
        for slice in slice_insert:
            self.logger.info("Executing inserts in %s.%s. Slice %s. Rows per slice %s." %  (loading_schema, table, num_insert, copy_limit ,   ))
            offset = slice*copy_limit
            sql_fallback = "SELECT %s FROM `%s`.`%s` LIMIT %s, %s;" % (select_stat, schema, table, offset, copy_limit)
            self.cursor_unbuffered.execute(sql_fallback)
            insert_data =  self.cursor_unbuffered.fetchall()
            self.pg_engine.insert_data(loading_schema, table, insert_data , column_list)
            num_insert +=1


    def print_progress (self, iteration, total, schema, table):
        """
            Print the copy progress in slices and estimated total slices.
            In order to reduce noise when the log level is info only the tables copied in multiple slices
            get the print progress.

            :param iteration: The slice number currently processed
            :param total: The estimated total slices
            :param table_name: The table name
        """
        if iteration>=total:
            total = iteration
        if total>1:
            self.logger.info("Table %s.%s copied %s slice of %s" % (schema, table, iteration, total))
        else:
            self.logger.debug("Table %s.%s copied %s slice of %s" % (schema, table, iteration, total))

    def __create_indices(self, schema, table):
        """
            The method copy the data between the origin and destination table.
            The method locks the table read only mode and  gets the log coordinates which are returned to the calling method.

            :param schema: the origin's schema
            :param table: the table name
            :return: the table and schema name with the primary key.
            :rtype: dictionary
        """
        loading_schema = self.schema_loading[schema]["loading"]
        self.connect_db_buffered()
        self.logger.debug("Creating indices on table %s.%s " % (schema, table))
        sql_index = """

            SELECT
            CASE WHEN index_name='PRIMARY'
            THEN
                index_name
            WHEN (
                    SELECT
                        count(1)
                    FROM
                        information_schema.statistics s
                    WHERE
                             s.index_name=t.index_name
                        AND  s.table_schema=t.table_schema
                )>1
            THEN
                concat(substring(index_name,1,59),'_',SUBSTRING(md5(uuid()),1,4))
            ELSE
                index_name
            END AS index_name,
            non_unique as non_unique,
            GROUP_CONCAT(column_name ORDER BY seq_in_index) as index_columns
            FROM
                information_schema.statistics t
            WHERE
                    table_schema=%s
                AND 	table_name=%s
                AND	index_type = 'BTREE'
            GROUP BY
                table_name,
                non_unique,
                index_name
            HAVING SUM(sub_part IS NOT NULL) = 0
            ;
        """

        self.cursor_buffered.execute(sql_index, (schema, table))
        index_data = self.cursor_buffered.fetchall()
        table_pkey = self.pg_engine.create_indices(loading_schema, table, index_data)
        self.disconnect_db_buffered()
        return table_pkey


    def __copy_tables(self):
        """
            The method copies the data between tables, from the mysql schema to the corresponding
            postgresql loading schema. Before the copy starts the table is locked and then the lock is released.
            If keep_existing_schema is true for the source then the tables are truncated before the copy,
            the indices are left in place and a REINDEX TABLE is executed after the copy.
        """


        for schema in self.schema_tables:
            loading_schema = self.schema_loading[schema]["loading"]
            destination_schema = self.schema_loading[schema]["destination"]
            table_list = self.schema_tables[schema]
            for table in table_list:
                self.logger.info("Copying the source table %s into %s.%s" %(table, loading_schema, table) )
                try:
                    if self.keep_existing_schema:
                        table_pkey = self.pg_engine.get_existing_pkey(destination_schema,table)
                        self.logger.info("Collecting constraints and indices from the destination table  %s.%s" %(destination_schema, table) )
                        self.pg_engine.collect_idx_cons(destination_schema,table)
                        self.logger.info("Removing constraints and indices from the destination table  %s.%s" %(destination_schema, table) )
                        self.pg_engine.cleanup_idx_cons(destination_schema,table)
                        self.logger.info("Truncating the table  %s.%s" %(destination_schema, table) )
                        self.pg_engine.truncate_table(destination_schema,table)
                        master_status = self.copy_data(schema, table)
                    else:
                        if self.copy_table_data:
                            master_status = self.copy_data(schema, table)
                        else:
                            master_status = self.get_master_coordinates()

                        table_pkey = self.__create_indices(schema, table)
                    self.pg_engine.store_table(destination_schema, table, table_pkey, master_status)
                    if self.keep_existing_schema:
                        #input("Press Enter to continue...")
                        self.logger.info("Adding constraint and indices to the destination table  %s.%s" %(destination_schema, table) )
                        self.pg_engine.create_idx_cons(destination_schema,table)
                except:
                    self.logger.info("Could not copy the table %s. Excluding it from the replica." %(table) )
                    raise

    def set_copy_max_memory(self):
        """
            The method sets the class variable self.copy_max_memory using the value stored in the
            source setting.

        """
        copy_max_memory = str(self.source_config["copy_max_memory"])[:-1]
        copy_scale = str(self.source_config["copy_max_memory"])[-1]
        try:
            int(copy_scale)
            copy_max_memory = self.source_config["copy_max_memory"]
        except:
            if copy_scale =='k':
                copy_max_memory = str(int(copy_max_memory)*1024)
            elif copy_scale =='M':
                copy_max_memory = str(int(copy_max_memory)*1024*1024)
            elif copy_scale =='G':
                copy_max_memory = str(int(copy_max_memory)*1024*1024*1024)
            else:
                print("**FATAL - invalid suffix in parameter copy_max_memory  (accepted values are (k)ilobytes, (M)egabytes, (G)igabytes.")
                sys.exit(3)
        self.copy_max_memory = copy_max_memory

    def __init_read_replica(self):
        """
            The method calls the pre-steps required by the read replica method.

        """

        self.replica_conn = {}
        self.source_config = self.sources[self.source]
        try:
            exit_on_error = True if self.source_config["on_error_read"]=='exit' else False
        except KeyError:
            exit_on_error = True
        self.my_server_id = self.source_config["my_server_id"]
        self.limit_tables = self.source_config["limit_tables"]
        self.skip_tables = self.source_config["skip_tables"]
        self.replica_batch_size = self.source_config["replica_batch_size"]
        self.sleep_loop = self.source_config["sleep_loop"]
        self.postgis_present = self.pg_engine.check_postgis()
        if self.postgis_present:
            self.hexify = self.hexify_always
        else:
            self.hexify = self.hexify_always + self.spatial_datatypes
        try:
            self.connect_db_buffered()
        except:
            if exit_on_error:
                raise
            else:
                return "skip"
        self.pg_engine.connect_db()
        self.schema_mappings = self.pg_engine.get_schema_mappings()
        self.schema_replica = [schema for schema in self.schema_mappings]
        db_conn = self.source_config["db_conn"]
        self.replica_conn["host"] = str(db_conn["host"])
        self.replica_conn["user"] = str(db_conn["user"])
        self.replica_conn["passwd"] = str(db_conn["password"])
        self.replica_conn["port"] = int(db_conn["port"])
        self.__build_table_exceptions()
        self.__build_skip_events()
        self.__check_mysql_config()
        if self.gtid_mode:
            master_data = self.get_master_coordinates()
            self.start_xid = master_data[0]["Executed_Gtid_Set"].split(':')[1].split('-')[0]

    def __init_sync(self):
        """
            The method calls the common steps required to initialise the database connections and
            class attributes within sync_tables,refresh_schema and init_replica.
        """
        try:
            self.source_config = self.sources[self.source]
        except KeyError:
            self.logger.error("The source %s doesn't exists " % (self.source))
            sys.exit()
        self.out_dir = self.source_config["out_dir"]
        self.copy_mode = self.source_config["copy_mode"]
        self.pg_engine.lock_timeout = self.source_config["lock_timeout"]
        self.pg_engine.grant_select_to = self.source_config["grant_select_to"]
        if "keep_existing_schema" in self.sources[self.source]:
            self.keep_existing_schema = self.sources[self.source]["keep_existing_schema"]
        else:
            self.keep_existing_schema = False
        self.set_copy_max_memory()
        self.postgis_present = self.pg_engine.check_postgis()
        if self.postgis_present:
            self.hexify = self.hexify_always
        else:
            self.hexify = self.hexify_always + self.spatial_datatypes
        self.connect_db_buffered()
        self.pg_engine.connect_db()
        self.schema_mappings = self.pg_engine.get_schema_mappings()
        self.pg_engine.schema_tables = self.schema_tables




    def refresh_schema(self):
        """
            The method performs a sync for an entire schema within a source.
            The method works in a similar way like init_replica.
            The swap happens in a single transaction.
        """
        self.logger.debug("starting sync schema for source %s" % self.source)
        self.logger.debug("The schema affected is %s" % self.schema)
        self.__init_sync()
        self.__check_mysql_config()
        self.pg_engine.set_source_status("syncing")
        self.__build_table_exceptions()
        self.schema_list = [self.schema]
        self.get_table_list()
        self.create_destination_schemas()
        try:
            self.pg_engine.schema_loading = self.schema_loading
            self.pg_engine.schema_tables = self.schema_tables
            if self.keep_existing_schema:
                self.disconnect_db_buffered()
                self.__copy_tables()
            else:
                self.create_destination_tables()
                self.disconnect_db_buffered()
                self.__copy_tables()
                self.pg_engine.grant_select()
                self.pg_engine.swap_schemas()
                self.drop_loading_schemas()
            self.pg_engine.set_source_status("initialised")
            self.connect_db_buffered()
            master_end = self.get_master_coordinates()
            self.disconnect_db_buffered()
            self.pg_engine.set_source_highwatermark(master_end, consistent=False)
            self.pg_engine.cleanup_table_events()
            notifier_message = "refresh schema %s for source %s is complete" % (self.schema, self.source)
            self.notifier.send_message(notifier_message, 'info')
            self.logger.info(notifier_message)
        except:
            if not self.keep_existing_schema:
                self.drop_loading_schemas()
            self.pg_engine.set_source_status("error")
            notifier_message = "refresh schema %s for source %s failed" % (self.schema, self.source)
            self.notifier.send_message(notifier_message, 'critical')
            self.logger.critical(notifier_message)
            raise



    def sync_tables(self):
        """
            The method performs a sync for specific tables.
            The method works in a similar way like init_replica except when swapping the relations.
            The tables are loaded into a temporary schema and the log coordinates are stored with the table
            in the replica catalogue. When the load is complete the method drops the existing table and changes the
            schema for the loaded tables to the destination schema.
            The swap happens in a single transaction.
        """
        self.logger.info("Starting sync tables for source %s" % self.source)
        self.__init_sync()
        self.__check_mysql_config()
        self.pg_engine.set_source_status("syncing")
        if self.tables == 'disabled':
            self.tables = self.pg_engine.get_tables_disabled ()
            if not self.tables:
                self.logger.info("There are no disabled tables to sync")
                return
        self.logger.debug("The tables affected are %s" % self.tables)
        self.__build_table_exceptions()
        self.schema_list = [schema for schema in self.schema_mappings if schema in self.schema_only]
        self.get_table_list()
        self.create_destination_schemas()
        try:
            self.pg_engine.schema_loading = self.schema_loading
            self.pg_engine.schema_tables = self.schema_tables
            if self.keep_existing_schema:
                self.disconnect_db_buffered()
                self.__copy_tables()
            else:
                self.create_destination_tables()
                self.disconnect_db_buffered()
                self.__copy_tables()
                self.pg_engine.grant_select()
                self.pg_engine.swap_tables()
                self.drop_loading_schemas()
            self.pg_engine.set_source_status("synced")
            self.connect_db_buffered()
            master_end = self.get_master_coordinates()
            self.disconnect_db_buffered()
            self.pg_engine.set_source_highwatermark(master_end, consistent=False)
            self.pg_engine.cleanup_table_events()
            notifier_message = "the sync for tables %s in source %s is complete" % (self.tables, self.source)
            self.notifier.send_message(notifier_message, 'info')
            self.logger.info(notifier_message)
        except:
            if not self.keep_existing_schema:
                self.drop_loading_schemas()
            self.pg_engine.set_source_status("error")
            notifier_message = "the sync for tables %s in source %s failed" % (self.tables, self.source)
            self.notifier.send_message(notifier_message, 'critical')
            self.logger.critical(notifier_message)
            raise

    def __get_text_spatial(self,raw_data):
        """
            The method returns the text representation converted in postgresql format
            for the raw data point using the ST_AsText function and the regular expressions

            :param charset: The table's character set
            :param raw_data: The raw_data returned by the mysql-replication library
            :return: text representation converted in postgresql format
            :rtype: text
        """
        decoded_data=binascii.hexlify(raw_data)
        return decoded_data.decode()[8:]

    def get_table_type_map(self):
        """
            The method builds a dictionary with a key per each schema replicated.
            Each key maps a dictionary with the schema's tables stored as keys and the column/type mappings.
            The dictionary is used in the read_replica method, to determine whether a field requires hexadecimal conversion.
        """
        table_type_map = {}
        table_map = {}
        self.logger.debug("collecting table type map")
        for schema in self.schema_replica:
            sql_tables = """
                SELECT
                    t.table_schema as table_schema,
                    t.table_name as table_name,
                    SUBSTRING_INDEX(t.TABLE_COLLATION,'_',1) as character_set
                FROM
                    information_schema.TABLES t
                WHERE
                        table_type='BASE TABLE'
                    AND	table_schema=%s
                ;
            """
            self.cursor_buffered.execute(sql_tables, (schema, ))
            table_list = self.cursor_buffered.fetchall()

            for table in table_list:
                column_type = {}
                sql_columns = """
                    SELECT
                        column_name as column_name,
                        data_type as data_type
                    FROM
                        information_schema.COLUMNS
                    WHERE
                            table_schema=%s
                        AND table_name=%s
                    ORDER BY
                        ordinal_position
                    ;
                """
                table_charset = table["character_set"]
                self.cursor_buffered.execute(sql_columns, (table["table_schema"], table["table_name"]))
                column_data = self.cursor_buffered.fetchall()
                for column in column_data:
                    column_type[column["column_name"]] = column["data_type"]
                table_dict = {}
                table_dict["table_charset"] = table_charset
                table_dict["column_type"] = column_type
                table_map[table["table_name"]] = table_dict
            table_type_map[schema] = table_map
            table_map = {}
        return table_type_map


    def __store_binlog_event(self, table, schema):
        """
        The private method returns whether the table event should be stored or not in the postgresql log replica.

        :param table: The table's name to check
        :param schema: The table's schema name
        :return: true if the table should be replicated, false if shouldn't
        :rtype: boolean
        """
        if self.tables_disabled:
            if  "%s.%s" % (schema, table) in self.tables_disabled:
                return False

        if schema in self.skip_tables:
            if table in self.skip_tables[schema]:
                return False

        if schema in self.limit_tables:
            if table in self.limit_tables[schema]:
                return True
            else:
                return False

        return True


    def __skip_event(self, table, schema, binlogevent):
        """
            The method returns true or false if whether the event should be skipped or not.
            The dictionary self.skip_events is used for the check.

            :param table: The table's name to check
            :param schema: The table's schema name
            :param binlogevent: The binlog event to evaluate
            :return: list with first element a boolean and the second element the event type
            :rtype: listt
        """
        if isinstance(binlogevent, DeleteRowsEvent):
            event = "delete"
        elif isinstance(binlogevent, UpdateRowsEvent):
            event = "update"
        elif isinstance(binlogevent, WriteRowsEvent):
            event = "insert"

        skip_event = False

        if self.skip_events:
            if self.skip_events[event]:
                table_name = "%s.%s" % (schema, table)
                if schema in self.skip_events[event] or table_name in self.skip_events[event]:
                    skip_event = True

        return [skip_event, event]

    def __build_gtid_set(self, gtid):
        """
            The method builds a gtid set using the current gtid and
        """
        new_set = None
        gtid_pack = []
        master_data= self.get_master_coordinates()
        if "Executed_Gtid_Set" in master_data[0]:
            gtid_set = master_data[0]["Executed_Gtid_Set"]
            gtid_list = gtid_set.split(",\n")
            for gtid_item in gtid_list:
                if gtid_item.split(':')[0] in gtid:
                    gtid_old = gtid_item.split(':')
                    gtid_new = "%s:%s-%s" % (gtid_old[0],gtid_old[1].split('-')[0],gtid[gtid_old[0]])
                    gtid_pack.append(gtid_new)
                else:
                    gtid_pack.append(gtid_item)
            new_set = ",\n".join(gtid_pack)
        return new_set

    def __decode_dic_keys(self, dic_encoded):
        """
        Private method to recursively decode the dictionary keys  and values into strings.
        This is used fixing the the json data types in the __read_replica_stream method because
        at moment the mysql-replication library returns the keys of the json data types as binary values in python3.

        :param dic_encoded: The dictionary with the encoded keys
        :return: The dictionary with the decoded keys
        :rtype: dictionary
        """
        dic_decoded = {}
        lst_decode = []
        if isinstance(dic_encoded, list):
            for item in dic_encoded:
                lst_decode.append(self.__decode_dic_keys(item))
            return lst_decode
        elif not isinstance(dic_encoded, dict):
            try:
                return dic_encoded.decode("UTF-8")
            except AttributeError:
                return dic_encoded
        else:
            for key, value in dic_encoded.items():
                try:
                    dic_decoded[key.decode("UTF-8")] = self.__decode_dic_keys(value)
                except AttributeError:
                    dic_decoded[key] = self.__decode_dic_keys(value)
        return dic_decoded

    def __read_replica_stream(self, batch_data):
        """
        Stream the replica using the batch data. This method evaluates the different events streamed from MySQL
        and manages them accordingly. The BinLogStreamReader function is called with the only_event parameter which
        restricts the event type received by the streamer.
        The events managed are the following.
        RotateEvent which happens whether mysql restarts or the binary log file changes.
        QueryEvent which happens when a new row image comes in (BEGIN statement) or a DDL is executed.
        The BEGIN is always skipped. The DDL is parsed using the sql_token class.
        [Write,Update,Delete]RowEvents are the row images pulled from the mysql replica.

        The RotateEvent and the QueryEvent cause the batch to be closed.

        The for loop reads the row events, builds the dictionary carrying informations like the destination schema,
        the 	binlog coordinates and store them into the group_insert list.
        When the number of events exceeds the replica_batch_size the group_insert is written into PostgreSQL.
        The batch is not closed in that case and the method exits only if there are no more rows available in the stream.
        Therefore the replica_batch_size is just the maximum size of the single insert and the size of replayed batch on PostgreSQL.
        The binlog switch or a captured DDL determines whether a batch is closed and processed.

        The update row event stores in a separate key event_before the row image before the update. This is required
        to allow updates where the primary key is updated as well.

        Each row event is scanned for data types requiring conversion to hex string.

        :param batch_data: The list with the master's batch data.
        :return: the batch's data composed by binlog name, binlog position and last event timestamp read from the mysql replica stream.
        :rtype: dictionary
        """
        size_insert=0
        sql_tokeniser = sql_token()
        table_type_map = self.get_table_type_map()
        inc_tables = self.pg_engine.get_inconsistent_tables()
        self.tables_disabled = self.pg_engine.get_tables_disabled(format='list')
        close_batch = False
        master_data = {}
        group_insert = []
        next_gtid = {}
        id_batch = batch_data[0][0]
        log_file = batch_data[0][1]
        log_position = batch_data[0][2]
        log_table = batch_data[0][3]
        master_data["log_table"] = log_table
        if self.gtid_mode:
            gtid_position = batch_data[0][4]
            gtid_pack = gtid_position.split(",\n")
            for gtid in gtid_pack:
                gtid  = gtid.split(':')
                next_gtid[gtid [0]]  = gtid [1].split("-")[-1]
                gtid_set = self.__build_gtid_set(next_gtid)
        else:
            gtid_set = None
        stream_connected = False
        my_stream = BinLogStreamReader(
            connection_settings = self.replica_conn,
            server_id = self.my_server_id,
            only_events = [RotateEvent, DeleteRowsEvent, WriteRowsEvent, UpdateRowsEvent, QueryEvent, GtidEvent, HeartbeatLogEvent],
            log_file = log_file,
            log_pos = log_position,
            auto_position = gtid_set,
            resume_stream = True,
            only_schemas = self.schema_replica,
            slave_heartbeat = self.sleep_loop,

        )
        if gtid_set:
            self.logger.debug("GTID ENABLED - gtid: %s. id_batch: %s " % (gtid_set, id_batch))
        else:
            self.logger.debug("GTID DISABLED - log_file %s, log_position %s. id_batch: %s " % (log_file, log_position, id_batch))

        for binlogevent in my_stream:
            if isinstance(binlogevent, GtidEvent):
                if close_batch:
                    break
                gtid  = binlogevent.gtid.split(':')
                next_gtid[gtid [0]]  = gtid [1]
                master_data["gtid"] = next_gtid

            elif isinstance(binlogevent, RotateEvent):
                event_time = binlogevent.timestamp
                binlogfile = binlogevent.next_binlog
                position = binlogevent.position
                self.logger.debug("ROTATE EVENT - binlogfile %s, position %s. " % (binlogfile, position))
                if (log_file != binlogfile and stream_connected) or len(group_insert)>0:
                    close_batch = True
                master_data["File"]=binlogfile
                master_data["Position"]=position
                master_data["Time"]=event_time
                master_data["gtid"] = next_gtid
                stream_connected = True
                if close_batch:
                    break

            elif isinstance(binlogevent, HeartbeatLogEvent):
                self.logger.debug("HEARTBEAT EVENT - binlogfile %s " % (binlogevent.ident,))
                if len(group_insert)>0 or log_file != binlogevent.ident:
                    self.logger.debug("WRITING ROWS - binlogfile %s " % (binlogevent.ident,))
                    master_data["File"] = binlogevent.ident
                    close_batch = True
                    break

            elif isinstance(binlogevent, QueryEvent):
                event_time = binlogevent.timestamp
                try:
                    schema_query = binlogevent.schema.decode()
                except:
                    schema_query = binlogevent.schema

                if binlogevent.query.strip().upper() not in self.statement_skip and schema_query in self.schema_mappings:
                    close_batch=True
                    destination_schema = self.schema_mappings[schema_query]
                    log_position = binlogevent.packet.log_pos
                    master_data["File"] = binlogfile
                    master_data["Position"] = log_position
                    master_data["Time"] = event_time
                    master_data["gtid"] = next_gtid
                    if len(group_insert)>0:
                        self.pg_engine.write_batch(group_insert)
                        group_insert=[]
                    self.logger.info("QUERY EVENT - binlogfile %s, position %s.\n--------\n%s\n-------- " % (binlogfile, log_position, binlogevent.query))
                    sql_tokeniser.parse_sql(binlogevent.query)
                    for token in sql_tokeniser.tokenised:
                        write_ddl = True
                        table_name = token["name"]
                        store_query = self.__store_binlog_event(table_name, schema_query)
                        if store_query:
                            table_key_dic = "%s.%s" % (destination_schema, table_name)
                            if table_key_dic in inc_tables:
                                write_ddl = False
                                log_seq = int(log_file.split('.')[1])
                                log_pos = int(log_position)
                                table_dic = inc_tables[table_key_dic]
                                if log_seq > table_dic["log_seq"]:
                                    write_ddl = True
                                elif log_seq == table_dic["log_seq"] and log_pos >= table_dic["log_pos"]:
                                    write_ddl = True
                                if write_ddl:
                                    self.logger.info("CONSISTENT POINT FOR TABLE %s REACHED  - binlogfile %s, position %s" % (table_key_dic, binlogfile, log_position))
                                    self.pg_engine.set_consistent_table(table_name, destination_schema)
                                    inc_tables = self.pg_engine.get_inconsistent_tables()
                            if write_ddl:
                                event_time = binlogevent.timestamp
                                self.logger.debug("TOKEN: %s" % (token))

                                if len(token)>0:
                                    query_data={
                                        "binlog":log_file,
                                        "logpos":log_position,
                                        "schema": destination_schema,
                                        "batch_id":id_batch,
                                        "log_table":log_table
                                    }
                                    self.pg_engine.write_ddl(token, query_data, destination_schema)



                    sql_tokeniser.reset_lists()
                if close_batch:
                    if len(group_insert) > 0:
                        self.logger.debug("writing the remaining %s row events when the statement event occurs" % (len(group_insert),))
                        self.pg_engine.write_batch(group_insert)
                    my_stream.close()
                    return [master_data, close_batch]
            else:

                for row in binlogevent.rows:
                    event_after={}
                    event_before={}
                    event_insert = {}
                    add_row = True
                    log_file=binlogfile
                    log_position=binlogevent.packet.log_pos
                    table_name=binlogevent.table
                    event_time=binlogevent.timestamp
                    schema_row = binlogevent.schema
                    destination_schema = self.schema_mappings[schema_row]
                    table_key_dic = "%s.%s" % (destination_schema, table_name)
                    store_row = self.__store_binlog_event(table_name, schema_row)
                    skip_event = self.__skip_event(table_name, schema_row, binlogevent)
                    if store_row and not skip_event[0]:
                        if table_key_dic in inc_tables:
                            table_consistent = False
                            log_seq = int(log_file.split('.')[1])
                            log_pos = int(log_position)
                            table_dic = inc_tables[table_key_dic]
                            if log_seq > table_dic["log_seq"]:
                                table_consistent = True
                            elif log_seq == table_dic["log_seq"] and log_pos >= table_dic["log_pos"]:
                                table_consistent = True
                                self.logger.info("CONSISTENT POINT FOR TABLE %s REACHED  - binlogfile %s, position %s" % (table_key_dic, binlogfile, log_position))
                            if table_consistent:
                                add_row = True
                                self.pg_engine.set_consistent_table(table_name, destination_schema)
                                inc_tables = self.pg_engine.get_inconsistent_tables()
                            else:
                                add_row = False
                        column_map = table_type_map[schema_row][table_name]["column_type"]
                        table_charset = table_type_map[schema_row][table_name]["table_charset"]

                        global_data={
                                            "binlog":log_file,
                                            "logpos":log_position,
                                            "schema": destination_schema,
                                            "table": table_name,
                                            "batch_id":id_batch,
                                            "log_table":log_table,
                                            "event_time":event_time
                                        }
                        if add_row:
                            if skip_event[1] == "delete":
                                global_data["action"] = "delete"
                                event_after=row["values"]
                            elif skip_event[1] == "update":
                                global_data["action"] = "update"
                                event_after=row["after_values"]
                                event_before=row["before_values"]
                            elif skip_event[1] == "insert":
                                global_data["action"] = "insert"
                                event_after=row["values"]

                            for column_name in event_after:
                                try:
                                    column_type=column_map[column_name]
                                except KeyError:
                                    self.logger.debug("Detected inconsistent structure for the table  %s. The replay may fail. " % (table_name))
                                    column_type = 'text'
                                if column_type in self.hexify and event_after[column_name]:
                                    event_after[column_name]=binascii.hexlify(event_after[column_name]).decode()
                                elif column_type in self.hexify and isinstance(event_after[column_name], bytes):
                                    event_after[column_name] = ''
                                elif column_type == 'json':
                                    event_after[column_name] = self.__decode_dic_keys(event_after[column_name])
                                elif column_type in self.spatial_datatypes and event_after[column_name]:
                                    event_after[column_name] = self.__get_text_spatial(event_after[column_name])


                            for column_name in event_before:
                                try:
                                    column_type=column_map[column_name]
                                except KeyError:
                                    self.logger.debug("Detected inconsistent structure for the table  %s. The replay may fail. " % (table_name))
                                    column_type = 'text'
                                if column_type in self.hexify and event_before[column_name]:
                                    event_before[column_name]=binascii.hexlify(event_before[column_name]).decode()
                                elif column_type in self.hexify and isinstance(event_before[column_name], bytes):
                                    event_before[column_name] = ''
                                elif column_type == 'json':
                                    event_before[column_name] = self.__decode_dic_keys(event_after[column_name])
                                elif column_type in self.spatial_datatypes and event_after[column_name]:
                                    event_before[column_name] = self.__get_text_spatial(event_before[column_name])
                            event_insert={"global_data":global_data,"event_after":event_after,  "event_before":event_before}
                            size_insert += len(str(event_insert))
                            group_insert.append(event_insert)

                        master_data["File"]=log_file
                        master_data["Position"]=log_position
                        master_data["Time"]=event_time
                        master_data["gtid"] = next_gtid

                        if len(group_insert)>=self.replica_batch_size:

                            self.logger.info("Max rows per batch reached. Writing %s. rows. Size in bytes: %s " % (len(group_insert), size_insert))
                            self.logger.debug("Master coordinates: %s" % (master_data, ))
                            self.pg_engine.write_batch(group_insert)
                            size_insert=0
                            group_insert=[]
                            close_batch=True



        my_stream.close()
        if len(group_insert)>0:
            self.logger.debug("writing the last %s events" % (len(group_insert), ))
            self.pg_engine.write_batch(group_insert)
            close_batch=True

        return [master_data, close_batch]


    def read_replica(self):
        """
            The method gets the batch data from PostgreSQL.
            If the batch data is not empty then method read_replica_stream is executed to get the rows from
            the mysql replica stored into the PostgreSQL database.
            When the method exits the replica_data list is decomposed in the master_data (log name, position and last event's timestamp).
            If the flag close_batch is set then the master status is saved in PostgreSQL the batch id  returned by the method is
            is saved in the class variable id_batch.
            This variable is used to determine whether the old batch should be closed or not.
            If the variable is not empty then the previous batch gets closed with a simple update of the processed flag.

        """

        skip = self.__init_read_replica()
        if skip:
            self.logger.warning("Couldn't connect to the source database for reading the replica. Ignoring.")
        else:
            self.pg_engine.set_source_status("running")
            replica_paused = self.pg_engine.get_replica_paused()
            if replica_paused:
                self.logger.info("Read replica is paused")
                self.pg_engine.set_read_paused(True)
            else:
                self.pg_engine.set_read_paused(False)
                batch_data = self.pg_engine.get_batch_data()
                if len(batch_data)>0:
                    id_batch=batch_data[0][0]
                    self.logger.debug("Batch data %s " % (batch_data))
                    replica_data=self.__read_replica_stream(batch_data)
                    master_data=replica_data[0]
                    close_batch=replica_data[1]
                    if "gtid" in master_data:
                        master_data["Executed_Gtid_Set"] = self.__build_gtid_set(master_data["gtid"])
                    else:
                        master_data["Executed_Gtid_Set"] = ""
                    if close_batch:
                        self.master_status=[master_data]
                        self.logger.debug("trying to save the master data...")
                        next_id_batch=self.pg_engine.save_master_status(self.master_status)
                        if next_id_batch:
                            self.logger.debug("new batch created, saving id_batch %s in class variable" % (id_batch))
                            self.id_batch=id_batch
                        else:
                            self.logger.debug("batch not saved. using old id_batch %s" % (self.id_batch))
                        if self.id_batch:
                            self.logger.debug("updating processed flag for id_batch %s", (id_batch))
                            self.pg_engine.set_batch_processed(id_batch)
                            self.id_batch=None
                self.pg_engine.keep_existing_schema = self.keep_existing_schema
                self.pg_engine.check_source_consistent()


            self.disconnect_db_buffered()


    def init_replica(self):
        """
            The method performs a full init replica for the given source
        """
        self.logger.debug("starting init replica for source %s" % self.source)
        self.__init_sync()
        self.__check_mysql_config()
        master_start = self.get_master_coordinates()
        self.pg_engine.set_source_status("initialising")
        self.pg_engine.cleanup_source_tables()
        self.schema_list = [schema for schema in self.schema_mappings]
        self.__build_table_exceptions()
        self.get_table_list()
        self.create_destination_schemas()
        try:
            self.pg_engine.insert_source_timings()
            self.pg_engine.schema_loading = self.schema_loading
            if self.keep_existing_schema:
                self.disconnect_db_buffered()
                self.__copy_tables()
            else:
                self.create_destination_tables()
                self.disconnect_db_buffered()
                self.__copy_tables()
                self.pg_engine.grant_select()
                self.pg_engine.swap_schemas()
                self.drop_loading_schemas()
            self.pg_engine.clean_batch_data()
            self.pg_engine.save_master_status(master_start)
            self.pg_engine.set_source_status("initialised")
            self.connect_db_buffered()
            master_end = self.get_master_coordinates()
            self.disconnect_db_buffered()
            self.pg_engine.set_source_highwatermark(master_end, consistent=False)
            notifier_message = "init replica for source %s is complete" % self.source
            self.notifier.send_message(notifier_message, 'info')
            self.logger.info(notifier_message)

        except:
            if not self.keep_existing_schema:
                self.drop_loading_schemas()
            self.pg_engine.set_source_status("error")
            notifier_message = "init replica for source %s failed" % self.source
            self.logger.critical(notifier_message)
            self.notifier.send_message(notifier_message, 'critical')
            raise



import re

from parsy import alt, any_char, digit, eof, forward_declaration, seq, string, success, regex, whitespace


def ci_string(s):
    """
    This function creates a case-insensitive parser for a string

    :param s: the string to make a case-insensitive parser for.
    :return: case-insensitive parser for string s
    :rtype: parsy.Parser
    """
    return string(s.upper(), transform=lambda x: x.upper()).result(s)


def optional_space_around(p):
    """
    This function extends an existing parser with optional whitespace
    around it. Whitespace is stripped from the parser's result.

    :param p: the parser to extend
    :return: a new parser with optional whitespace around it
    :rtype: parsy.Parser
    """
    return whitespace.optional() >> p << whitespace.optional()


def parentheses_around(p):
    """
    This function extends an existing parser with parentheses
    around it. The captured parentheses and adjacent whitespace
    are discarded.

    :param p: the parser to extend
    :return: a new parser with parenthesis around the old parser
    :rtype: parsy.Parser
    """
    return optional_space_around(lparen) >> p << optional_space_around(rparen)


pgsql_identifier = regex(r"\w+")
mysql_identifier = string("`") >> regex(r"[\w\s]+") << string("`")
identifier = pgsql_identifier | mysql_identifier
number = digit.many().concat()

lparen = string("(")
rparen = string(")")
semicolon = string(";")
comma_sep = optional_space_around(string(","))
sql_string = (
    # single-quoted string
    (string("'") >> (string(r"\'") | any_char).until(string("'")).concat() << string("'")) |
    # double-quoted string
    (string('"') >> (string(r'\"') | any_char).until(string('"')).concat() << string('"'))
)

ci_word = regex(r"\w+")

class sql_token(object):
    """
    The class tokenises the sql statements captured by mysql_engine.
    Several regular expressions analyse and build the elements of the token.
    The DDL support is purposely limited to the following.

    DROP PRIMARY KEY
    CREATE (UNIQUE) INDEX/KEY
    CREATE TABLE
    ALTER TABLE

    The method post_process_key_definition ignores any foreign key definition from the
    sql statement as we don't enforce any foreign key on the PostgreSQL replication.
    """

    # for parsing complex index definitions
    # forward_declaration is like lazy initialisation
    inline_expr = forward_declaration()
    simple_expr = regex(r"[^\(\)]+")
    group_expr = seq(optional_space_around(lparen), inline_expr.many().concat(), optional_space_around(rparen)).concat()
    inline_expr.become(group_expr | simple_expr)

    # { column_name | column_name(n) | (FUNCTION(column_name)) }
    key_part = alt(
        # column_name(80)
        seq(
            identifier,
            optional_space_around(lparen), number, optional_space_around(rparen),
        ).concat().tag("partial"),

        # column_name
        identifier.tag("column"),

        # functional
        inline_expr.tag("functional"),
    )

    key_part_group = key_part.sep_by(comma_sep).map(
        lambda tagged_list: {
            "tags": [tag for (tag, _value) in tagged_list],
            "values": [value for (_tag, value) in tagged_list],
        }
    ).combine_dict(
        lambda tags, values: {
            "is_functional": "functional" in tags,
            "is_partial": "partial" in tags,
            "index_columns": values,
        }
    )

    # [CONSTRAINT pk_id] PRIMARY KEY (column_name, ...)
    pk_definition = seq(
        __constraint=(ci_string("CONSTRAINT") >> whitespace >> identifier >> whitespace).optional(),
        index_name=seq(ci_string("PRIMARY"), whitespace, ci_string("KEY")).result("PRIMARY"),
        index_columns=parentheses_around(key_part.sep_by(comma_sep)),
        non_unique=success(0),
        is_fulltext=success(False),
        is_spatial=success(False),
    ).combine_dict(dict)

    # [CONSTRAINT uk_xyz] UNIQUE [{KEY | INDEX}] (column_name, ...)
    uk_definition = seq(
        __constraint=(ci_string("CONSTRAINT") >> whitespace >> identifier >> whitespace).optional(),
        index_name=ci_string("UNIQUE").result("UNIQUE"),
        __key_or_index_keyword=(whitespace >> (ci_string("INDEX") | ci_string("KEY"))).optional(),
        __optional_idx_name=(whitespace >> identifier).optional(),
        index_type=(
            whitespace >> ci_string("USING") >> (ci_string("BTREE") | ci_string("HASH"))
        ).optional(),
        index_columns=parentheses_around(key_part.sep_by(comma_sep)),
        non_unique=success(0),
        is_fulltext=success(False),
        is_spatial=success(False),
    ).combine_dict(dict)

    # {INDEX | KEY} [idx_name] (column_name, ...)
    idx_definition = seq(
        index_name=(ci_string("INDEX") | ci_string("KEY")).result("INDEX"),
        __optional_idx_name=(whitespace >> identifier).optional(),
        index_type=(
            whitespace >> ci_string("USING") >> (ci_string("BTREE") | ci_string("HASH"))
        ).optional(),
        index_columns=parentheses_around(key_part.sep_by(comma_sep)),
        non_unique=success(1),
        is_fulltext=success(False),
        is_spatial=success(False),
    )

    # [CONSTRAINT fk_id] FOREIGN KEY REFERENCES table_name (column_name, ...)
    # [ON {UPDATE | DELETE} {CASCADE | RESTRICT}]
    fkey_definition = seq(
        __constraint=(ci_string("CONSTRAINT") >> whitespace >> identifier >> whitespace).optional(),
        index_name=(ci_string("FOREIGN") >> whitespace >> ci_string("KEY")).result("FOREIGN"),
        index_columns=parentheses_around(key_part.sep_by(comma_sep)),
        __references=whitespace.optional() >> ci_string("REFERENCES"),
        __other_table_name=whitespace.optional() >> identifier,
        __other_columns=parentheses_around(key_part.sep_by(comma_sep)),
        __on_delete_or_update=seq(
            whitespace, ci_string("ON"), whitespace, ci_string("UPDATE") | ci_string("DELETE"),
            whitespace, ci_string("CASCADE") | ci_string("RESTRICT"),
        ).many(),
        non_unique=success(1),
        is_fulltext=success(False),
        is_spatial=success(False),
    ).combine_dict(dict)

    # {SPATIAL | FULLTEXT} {INDEX | KEY} idx_name (column_name, ...)
    other_key_definition = seq(
        __constraint=(ci_string("CONSTRAINT") >> whitespace >> identifier >> whitespace).optional(),
        index_name=whitespace.optional() >> (ci_string("SPATIAL") | ci_string("FULLTEXT")),
        __index_or_key=(whitespace >> (ci_string("INDEX") | ci_string("KEY"))).optional(),
        __idx_name=whitespace >> identifier,
        index_columns=parentheses_around(key_part.sep_by(comma_sep)),
        non_unique=success(1),
    ).combine_dict(
        lambda index_name, **rest: {
            "is_spatial": index_name.upper() == "SPATIAL",
            "is_fulltext": index_name.upper() == "FULLTEXT",
            "index_name": "OTHER",
            **rest,
        }
    )

    any_key_definition = alt(
        pk_definition,
        uk_definition,
        idx_definition,
        fkey_definition,
        other_key_definition
    ).combine_dict(
        lambda index_columns, **kwargs: {
            "key_part_tags": [tag for (tag, _value) in index_columns],
            "index_columns": [value for (_tag, value) in index_columns],
            **kwargs,
        }
    ).combine_dict(
        lambda key_part_tags, **kwargs: {
            "is_functional": "functional" in key_part_tags,
            "is_partial": "partial" in key_part_tags,
            **kwargs,
        }
    )

    # column_name type [PRECISION | VARYING] [(numeric_dimension, ...)] [('enum_list', ...)]
    # [NOT NULL] [PRIMARY KEY] [DEFAULT {'string' | literal} ] [ignored extras ...]
    column_definition = seq(
        column_name=identifier,
        data_type=whitespace >> ci_word.map(lambda x: x.lower()),
        __precision_or_varying=(
            whitespace >>
            (ci_string("PRECISION") | ci_string("VARYING"))
        ).optional(),
        dimensions=(
            optional_space_around(lparen) >>
            number.sep_by(
                comma_sep | optional_space_around(string("|"))
            )
            << optional_space_around(rparen)
        ).optional(),
        enum_list=(
            optional_space_around(lparen) >>
            sql_string.sep_by(
                comma_sep | optional_space_around(string("|"))
            )
            << optional_space_around(rparen)
        ).optional(),
        extras=(
            whitespace.optional() >> alt(
                seq(ci_string("NOT"), whitespace, ci_string("NULL")).result("NOT NULL"),
                seq(ci_string("PRIMARY"), whitespace, ci_string("KEY")).result("PRIMARY KEY"),
                seq(
                    ci_string("DEFAULT").result("DEFAULT"),
                    whitespace >> (sql_string.map(lambda x: f"'{x}'") | ci_word)
                ),
                seq(ci_string("COMMENT"), whitespace, sql_string).result("COMMENT"),
                identifier,
                ci_word,
            ).sep_by(whitespace)
        ).optional(default=[]),
    ).combine_dict(dict)

    # CREATE TABLE [IF NOT EXISTS] table_name (key_definition | column_definition, ...) [anything extra]
    create_table_statement = seq(
        command=(ci_string("CREATE") >> whitespace >> ci_string("TABLE")).result("CREATE TABLE"),
        __if_not_exists=seq(
            whitespace, ci_string("IF"),
            whitespace, ci_string("NOT"),
            whitespace, ci_string("EXISTS")
        ).optional(),
        name=whitespace >> identifier,
        inner=whitespace.optional() >> lparen >> (
            optional_space_around(
                any_key_definition.tag("index") | column_definition.tag("column")
            ).sep_by(string(","))
        ) << rparen,
        __rest=any_char.many(),
    ).combine_dict(dict)

    # ALTER TABLE table_name RENAME [TO] new_name
    alter_rename_table_statement = seq(
        command=seq(ci_string("ALTER"), whitespace, ci_string("TABLE")).result("RENAME TABLE"),
        name=whitespace >> identifier,
        __rename=whitespace >> ci_string("RENAME"),
        __to=(whitespace >> ci_string("TO")).optional(),
        new_name=whitespace >> identifier,
    ).combine_dict(dict)

    # [schema_identifier.]table_name TO [new_schema_identifier.]new_name
    rename_table_item = seq(
        command=success("RENAME TABLE"),
        __from_schema=(identifier >> string(".")).optional(),
        name=identifier,
        __to=whitespace >> ci_string("TO"),
        __to_schema=whitespace >> (identifier >> string(".")).optional(),
        new_name=identifier,
    ).combine_dict(dict)

    # RENAME TABLE ( [old_schema.]old_name TO [new_schema.]new_name )
    # returns list, not dict!
    rename_table_statement = (
        ci_string("RENAME") >> whitespace >> ci_string("TABLE") >>
        optional_space_around(rename_table_item).sep_by(string(","))
    )

    # DROP TABLE [IF EXISTS] table_name
    drop_table_statement = seq(
        command=seq(ci_string("DROP"), whitespace, ci_string("TABLE")).result("DROP TABLE"),
        __if_exists=seq(whitespace, ci_string("IF"), whitespace, ci_string("EXISTS")).optional(),
        name=whitespace >> identifier,
    ).combine_dict(dict)

    # ALTER TABLE table_name DROP PRIMARY KEY
    drop_primary_key_statement = seq(
        command=seq(ci_string("ALTER"), whitespace, ci_string("TABLE")).result("DROP PRIMARY KEY"),
        name=whitespace >> identifier,
        __drop_pk=whitespace >> seq(
            ci_string("DROP"), whitespace, ci_string("PRIMARY"), whitespace, ci_string("KEY")
        )
    ).combine_dict(dict)

    # TRUNCATE [TABLE] [schema_name.]table_name
    truncate_table_statement = seq(
        command=ci_string("TRUNCATE"),
        __table=(whitespace >> ci_string("TABLE")).optional(),
        __space=whitespace,
        __schema=(identifier << string(".")).optional(),
        name=identifier,
    ).combine_dict(dict)

    # post processes a parsed column definition
    # when it occurs in an ALTER TABLE statement
    column_definition_in_alter_table = column_definition.combine_dict(
        lambda column_name, data_type, dimensions, enum_list, extras, **k: {
            "name": column_name,
            "type": data_type,
            "dimension": (
                ", ".join(dimensions) if dimensions else
                ", ".join(map(lambda x: f"'{x}'", enum_list)) if enum_list else
                0
            ),
            "default": (
                next(
                    (extra[1] for extra in extras
                    if isinstance(extra, list) and extra[0] == "DEFAULT"),
                    None
                )
            )
        }
    ).map(
        lambda d: dict(
            d,
            data_type=d["type"],
            column_type="%s(%s)" % (d["type"], d["dimension"]) if d["dimension"] else d["type"]
        )
    )

    # ADD [COLUMN] column_definition
    alter_table_add = seq(
        command=ci_string("ADD"),
        __column=(whitespace >> ci_string("COLUMN")).optional(),
        col_def=whitespace >> column_definition_in_alter_table
    ).combine_dict(
        lambda command, col_def: dict(command=command, **col_def)
    )

    # ADD [COLUMN] (column_definition, ...)
    alter_table_add_multiple = seq(
        command=ci_string("ADD").result("ADD MULTIPLE"),
        __column=(whitespace >> ci_string("COLUMN")).optional(),
        col_defs=optional_space_around(
            parentheses_around(
                column_definition_in_alter_table.sep_by(comma_sep)
            )
        ),
    ).combine_dict(dict)

    # CHANGE [COLUMN] column_name column_definition
    alter_table_change = seq(
        command=ci_string("CHANGE"),
        __column=(whitespace >> ci_string("COLUMN")).optional(),
        old=whitespace >> identifier,
        col_def=whitespace >> column_definition_in_alter_table,
    ).combine_dict(
        lambda command, old, col_def: dict(
            command=command,
            old=old,
            new=col_def["name"],
            **col_def,
        )
    ).combine_dict(dict)

    # DROP [COLUMN] column_name
    alter_table_drop = seq(
        command=ci_string("DROP"),
        __column=(whitespace >> ci_string("COLUMN")).optional(),
        name=whitespace >> identifier,
    ).combine_dict(dict)

    # MODIFY [COLUMN] column_definition
    alter_table_modify = seq(
        command=ci_string("MODIFY"),
        __column=(whitespace >> ci_string("COLUMN")).optional(),
        col_def=whitespace >> column_definition_in_alter_table,
    ).combine_dict(
        lambda command, col_def: dict(command=command, **col_def)
    )

    # ADD {INDEX | KEY} [index_name] [USING {BTREE | HASH}] (key_part, ...)
    alter_table_add_index = seq(
        command=ci_string("ADD").result("ADD INDEX"),
        key_dic=whitespace >> any_key_definition,
    ).combine_dict(
        lambda command, key_dic: dict(command=command, **key_dic)
    )

    # RENAME {INDEX | KEY} old_index_name TO new_index_name
    alter_table_rename_index = seq(
        command=seq(ci_string("RENAME"), whitespace, ci_string("INDEX") | ci_string("KEY")).result("RENAME INDEX"),
        index_name=whitespace >> identifier,
        __to=whitespace >> ci_string("TO"),
        new_index_name=whitespace >> identifier,
    ).combine_dict(dict)

    # DROP {INDEX | KEY} index_name
    alter_table_drop_index = seq(
        command=seq(ci_string("DROP"), whitespace, ci_string("INDEX") | ci_string("KEY")).result("DROP INDEX"),
        index_name=whitespace >> identifier,
    )

    # [ADD | DROP | CHANGE | MODIFY] {INDEX | KEY | CONSTRAINT | CHECK | UNIQUE | FOREIGN KEY | PRIMARY KEY}
    alter_table_ignored = seq(
        ci_string("ADD") | ci_string("DROP") | ci_string("CHANGE") | ci_string("MODIFY"),
        whitespace >> alt(
            ci_string("INDEX"), ci_string("KEY"), ci_string("CONSTRAINT"), ci_string("CHECK"),
            ci_string("UNIQUE"), seq(ci_string("FOREIGN"), whitespace, ci_string("KEY")),
            seq(ci_string("PRIMARY"), whitespace, ci_string("KEY"))
        ),
        any_char.until(string(",") | eof).concat(),
    ).result(None)

    # ALTER TABLE [schema_name.]table_name alter_table_subcommand, ...
    alter_table_statement = seq(
        command=seq(ci_string("ALTER"), whitespace, ci_string("TABLE")).result("ALTER TABLE"),
        __space=whitespace,
        __schema=seq(identifier, string(".")).optional(),
        name=identifier,
        alter_cmd=optional_space_around(
            alt(
                alter_table_add_index,
                alter_table_rename_index,
                alter_table_drop_index,
                alter_table_ignored,
                alter_table_drop,
                alter_table_add,
                alter_table_add_multiple,
                alter_table_change,
                alter_table_modify,
            ).sep_by(comma_sep, min=1).map(lambda ls: [x for x in ls if x])
        )
    ).combine_dict(dict)

    # CREATE [UNIQUE] INDEX index_name ON table_name (column_name, ...)
    create_index_statement = seq(
        command=ci_string("CREATE").result("ADD INDEX"),
        non_unique=(whitespace >> ci_string("UNIQUE")).result(0).optional(default=1),
        is_fulltext=(whitespace >> ci_string("FULLTEXT")).result(True).optional(default=False),
        is_spatial=(whitespace >> ci_string("SPATIAL")).result(True).optional(default=False),
        __index=whitespace >> ci_string("INDEX"),
        index_name=whitespace >> identifier,
        index_type=(
            optional_space_around(ci_string("USING")) >> (ci_string("BTREE") | ci_string("HASH"))
        ).optional(),
        __on=whitespace >> ci_string("ON"),
        __space=whitespace,
        __schema=seq(identifier, string(".")).optional(),
        name=identifier,
        key_parts=parentheses_around(key_part_group),
    ).combine_dict(
        lambda key_parts, **rest: {**key_parts, **rest}
    ).combine_dict(
        lambda command, name, **key_dic: {
            "command": "ALTER TABLE",
            "name": name,
            "alter_cmd": [{"command": command, **key_dic}],
        }
    )

    # DROP INDEX index_name ON table_name
    # equivalent to ALTER TABLE ... DROP INDEX ...
    drop_index_statement = seq(
        __drop_index=seq(ci_string("DROP"), whitespace, ci_string("INDEX")),
        index_name=whitespace >> identifier,
        __on=whitespace >> ci_string("ON"),
        name=whitespace >> identifier,
    ).combine_dict(
        lambda name, index_name: {
            "command": "ALTER TABLE",
            "name": name,
            "alter_cmd": [{"command": "DROP INDEX", "index_name": index_name}],
        }
    )

    def __init__(self):
        """
            Class constructor the regular expressions are compiled and the token lists are initialised.
        """
        self.tokenised = []
        self.query_list = []
        self.pkey_cols = []
        self.ukey_cols = []

        # supported statements
        # RENAME TABLE
        # CREATE TABLE
        # DROP TABLE
        # TRUNCATE TABLE
        # ALTER TABLE
        # CREATE INDEX
        # DROP INDEX
        self.sql_parser = optional_space_around(
            alt(
                self.rename_table_statement,
                self.create_table_statement.map(self._post_process_create_table),
                self.drop_table_statement,
                self.truncate_table_statement,
                self.drop_primary_key_statement,
                self.create_index_statement,
                self.drop_index_statement,
                self.alter_table_statement.map(self._post_process_alter_table),
                self.alter_rename_table_statement,
            ).optional()
        )

    def reset_lists(self):
        """
            The method resets the lists to empty lists after a successful tokenisation.
        """
        self.tokenised=[]
        self.query_list=[]

    def quote_cols(self, cols):
        """
            The method adds the " quotes to the column names.
            The string is converted to a list using the split method with the comma separator.
            The columns are then stripped and quoted with the "".
            Finally the list elements are rejoined in a string which is returned.
            The method is used in build_key_dic to sanitise the column names.

            :param cols: The columns string
            :return: The columns quoted between ".
            :rtype: text
        """
        idx_cols = cols.split(',')
        idx_cols = ['"%s"' % col.strip() for col in idx_cols]
        quoted_cols = ",".join(idx_cols)
        return quoted_cols

    def _post_process_key_definition(
        self,
        index_name,
        index_columns,
        non_unique,
        table_name,
        idx_counter,
        **tags,
    ):
        """
            This function builds a new key_dic by overwriting the index_name if necessary and by
            discarding indices that are to be ignored (foreign key, fulltext, etc.). This discarding
            is done by returning a None value instead of the key_dic.

            ```
            key_dic format:
                index_name: str
                index_columns: list[str]
                non_unique: int 0|1
            ```

            :param index_name: The kind of index. One of PRIMARY, UNIQUE, INDEX, FOREIGN, OTHER
            :param index_columns: The columns covered by this index
            :param non_unique: Whether this index must enforce unique check or not
            :param table_name: The name of the table that is used to create a new index name
            :param idx_counter: An index counter that is used to create a new index name
            :param tags: Flags such as is_functional, is_partial, is_fulltext, etc.
            :return: The transformed key dic or None
            :rtype: dictionary | None
        """
        if index_name in {"FOREIGN", "OTHER"}:
            return None
        elif index_name == "PRIMARY":
            return dict(
                index_name="PRIMARY",
                index_columns=index_columns,
                non_unique=0,
                **tags,
            )
        elif index_name == "UNIQUE":
            return dict(
                index_name=f"ukidx_{table_name[0:20]}_{idx_counter}",
                index_columns=index_columns,
                non_unique=0,
                **tags,
            )
        elif index_name == "INDEX":
            return dict(
                index_name=f"idx_{table_name[0:20]}_{idx_counter}",
                index_columns=index_columns,
                non_unique=1,
                **tags,
            )
        else:
            raise Exception(f"Unknown index name: {index_name}")

    def _post_process_column_definition(
            self, column_name, data_type, dimensions, enum_list, extras
    ):
        """
            This function does uses the parts identified by the column_definition parser
            and builds a dictionary in the col_dic format. It adds fields that are not identified
            directly by the parser.

            ```
            col_dic format:
              column_name: str
              data_type: str
              is_nullable: enum "YES"|"NO"
              enum_list: maybe str
              character_maximum_length: maybe str
              numeric_precision: maybe str
              numeric_scale: maybe str | int (defaults to 0)
              extra: str
              column_type: str
            ```

            The arguments accepted are the ones parsed by the column_definition parser.

            :param column_name: column_name as parsed by the column definition parser
            :param data_type: data_type as parsed by the column definition parser
            :param dimensions: the numeric dimensions defined as part of the column type
            :param enum_list: the enum list defined as part of the column type
            :param extras: other modifiers to the column
            :return: column dictionary which conforms to the col_dic format
            :rtype: dictionary
        """
        col_dict = dict(column_name=column_name, data_type=data_type)

        col_dict["is_nullable"] = "NO" if "NOT NULL" in extras else "YES"
        col_dict["extra"] = "auto_increment" if "AUTO_INCREMENT" in extras else ""

        if dimensions:
            col_dict["numeric_precision"] = col_dict["character_maximum_length"] = dimensions[0] # str
            col_dict["numeric_scale"] = dimensions[1] if len(dimensions) > 1 else 0 # str or int
        elif enum_list:
            col_dict["enum_list"] = "( %s )" % ", ".join(map(lambda x: f"'{x}'", enum_list))

        if dimensions:
            col_dict["column_type"] = "%s(%s)" % (data_type, ", ".join(dimensions))
        elif enum_list:
            col_dict["column_type"] = "%s(%s)" % (data_type, ", ".join(map(lambda x: f"'{x}'", enum_list)))
        else:
            col_dict["column_type"] = data_type

        return col_dict

    def _post_process_create_table(self, table_dic):
        """
            The method parse and generates a dictionary from the CREATE TABLE statement.

            The part of statement inside round brackets is parsed for column and index
            definitions. Index and column definitions are separated and processed one by one.
            First, indices are processed and added to self.pkey_cols and self.ukey_cols.
            Then columns are parsed and pkey_cols is modified if an inline primary key had been
            set.

            The indices are stored in the dictionary key "indices" as a list of dictionaries. Each
            key_dic has a fixed set of keys, as returned by _post_process_key_definition.

            The columns are stored in the dictionary key "columns" as a list of dictionaries. Each
            col_dic has a fixed set of keys, as returned by _post_process_column_definition.

            :param sql_create: The sql string with the CREATE TABLE statement
            :return: table_dic the table dictionary tokenised from the CREATE TABLE
            :rtype: dictionary
        """

        columns_and_indices = table_dic.pop("inner")

        columns, indices = [], []
        for col_or_idx in columns_and_indices:
            tag, value = col_or_idx
            if tag == "column":
                columns.append(value)
            elif tag == "index":
                indices.append(value)
            else:
                raise Exception(f"unknown tag: {tag}")

        table_dic["columns"], table_dic["indices"] = [], []

        # post-process indices
        for raw_key_dic in indices:
            key_dic = self._post_process_key_definition(
                **raw_key_dic,
                table_name=table_dic["name"],
                idx_counter=len(table_dic["indices"])
            )
            if key_dic:
                table_dic["indices"].append(key_dic)

                # update self.pkey_cols or self.ukey_cols
                if key_dic["index_name"] == "PRIMARY":
                    self.pkey_cols = list(key_dic["index_columns"])
                elif raw_key_dic["index_name"] == "UNIQUE":
                    self.ukey_cols += [col_name for col_name in key_dic["index_columns"]]

        # post-process columns
        for raw_col_dic in columns:
            col_dic = self._post_process_column_definition(**raw_col_dic)
            if col_dic:
                # check for inline primary key definition
                if "PRIMARY KEY" in raw_col_dic["extras"]:
                    table_dic["indices"].append(dict(
                        index_name="PRIMARY", index_columns=[col_dic["column_name"]], non_unique=0
                    ))
                    self.pkey_cols = [col_dic["column_name"]]

                # must be non-nullable if column is ukey or pkey
                if (col_dic["column_name"] in self.pkey_cols or
                        col_dic["column_name"] in self.ukey_cols):
                    col_dic["is_nullable"] = "NO"

                table_dic["columns"].append(col_dic)

        return table_dic

    def _post_process_alter_table(self, alter_table_dic):
        """
        Post process alter table statement's parsed result to make it
        compatible with the expected result type.
        """
        # make a copy of alter_cmd list so we can
        # modify the original list freely
        alter_cmds = list(alter_table_dic["alter_cmd"])

        for cmd in alter_cmds:
            if cmd["command"] == "ADD MULTIPLE":
                alter_table_dic["alter_cmd"].remove(cmd)
                alter_table_dic["alter_cmd"].extend([
                    dict(command="ADD", **col_def)
                    for col_def in cmd["col_defs"]
                ])
        return alter_table_dic

    def parse_sql(self, sql_string):
        """
            The method removes comments from the sql string and parses it.
            The statements are split in a list using the statement separator ;

            For each statement, the sql parser tries to parse it with
            a set of specific statement parsers. If any supported statement
            is found, it is parsed and returned as a non-empty dict or
            a list of dicts by the parser.

            Look at self.sql_parser for the different statements that are
            supported by the parser.

            Different match are performed on the statement.
            RENAME TABLE
            CREATE TABLE
            DROP TABLE
            ALTER TABLE
            ALTER INDEX
            DROP PRIMARY KEY
            TRUNCATE TABLE

            Each successful parse builds a dictionary with at least two
            keys "name" and "command".

            Each statement parse comes with specific addictional keys.

            When the token dictionary is complete is added to the class list tokenised

            :param sql_string: The sql string with the sql statements.
        """
        sql_string_cleanup = re.sub(r'/\*.*?\*/', '', sql_string, re.DOTALL)
        sql_string_cleanup = re.sub(r'--.*?\n', '', sql_string_cleanup)

        multiple_sql_parser = self.sql_parser.sep_by(optional_space_around(string(";")))
        for stat_dic in multiple_sql_parser.parse(sql_string_cleanup):
            if isinstance(stat_dic, dict) and stat_dic != {}:
                self.tokenised.append(stat_dic)
            elif isinstance(stat_dic, list):
                for d in stat_dic:
                    if isinstance(d, dict) and stat_dic != {}:
                        self.tokenised.append(stat_dic)



#!/usr/bin/env python
import time
import psycopg2
from psycopg2.extras import LogicalReplicationConnection
from psycopg2.extras import  REPLICATION_LOGICAL
strconn = "dbname=db_replica user=usr_replica host=foo password=bar port=5432"  
log_conn = psycopg2.connect(strconn, connection_factory=LogicalReplicationConnection)
log_cur = log_conn.cursor()
log_cur.create_replication_slot("logical1", slot_type=REPLICATION_LOGICAL, output_plugin="test_decoding")
#log_cur.start_replication(slot_name="logical1", slot_type=REPLICATION_LOGICAL, start_lsn=0, timeline=0, options=None, decode=True)
"""while True:
	log_cur.send_feedback()
	msg = log_cur.read_message()
	if msg:
		print(msg.payload)
	time.sleep(1)
"""
log_cur.drop_replication_slot("logical1")
log_conn.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import setuptools

def readme():
    with open('README.rst') as f:
        return f.read()

package_data = {'pg_chameleon': ['configuration/config-example.yml','sql/upgrade/*.sql','sql/drop_schema.sql','sql/create_schema.sql', 'LICENSE.txt']}

setuptools.setup(
     name="pg_chameleon",
     version="2.0.20",
     description="MySQL to PostgreSQL replica and migration",
    long_description=readme(),
    author = "Federico Campoli",
    author_email = "thedoctor@pgdba.org",
    maintainer = "Federico Campoli",
    maintainer_email = "thedoctor@pgdba.org",
    url="https://github.com/the4thdoctor/pg_chameleon/",
    license="BSD License",
    platforms=[
        "linux"
    ],
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Natural Language :: English",
        "Operating System :: POSIX :: BSD",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: Other/Nonlisted Topic"
    ],
    py_modules=[
        "pg_chameleon.__init__",
        "pg_chameleon.lib.global_lib",
        "pg_chameleon.lib.mysql_lib",
        "pg_chameleon.lib.pg_lib",
        "pg_chameleon.lib.sql_util"
    ],
    scripts=[
        "scripts/chameleon.py",
        "scripts/chameleon"
    ],
    install_requires=[
        'PyMySQL>=0.10.0',
        'mysql-replication>=0.31',
        'psycopg2-binary>=2.8.3',
        'PyYAML>=3.13',
        'tabulate>=0.8.1',
        'daemonize>=2.4.7',
        'rollbar>=0.13.17',
        'parsy>=2.1',
        'Sphinx>=7.4.7'

    ],
    include_package_data = True,
    package_data=package_data,
    packages=setuptools.find_packages(),
    python_requires='>=3.7',
    keywords='postgresql mysql replica migration database',

)



import rollbar
import pprint
import yaml
import os, os.path
import sys
import time
import signal
from shutil import copy
from distutils.sysconfig import get_python_lib
from tabulate import tabulate
from pg_chameleon import pg_engine, mysql_source, pgsql_source
import logging
from logging.handlers  import TimedRotatingFileHandler
from daemonize import Daemonize
import multiprocessing as mp
import traceback
from pkg_resources import get_distribution
class rollbar_notifier(object):
    """
        This class is used to send messages to rollbar whether the key and environment variables are set
    """
    def __init__(self, rollbar_key, rollbar_env, rollbar_level, logger):
        """
            Class constructor.
        """
        self.levels = {
                "critical": 1,
                "error": 2,
                "warning": 3,
                "info": 5
            }
        self.rollbar_level = self.levels[rollbar_level]
        self.logger = logger
        self.notifier = rollbar
        if rollbar_key !='' and rollbar_env != '':
            self.notifier.init(rollbar_key, rollbar_env)
        else:
            self.notifier = None


    def send_message(self, message, level):
        """
            The method sends a message to rollbar. If it fails it just logs an error
            without causing the process to crash.
        """
        if self.notifier:
            exc_info = sys.exc_info()
            try:
                notification_level = self.levels[level]
                if notification_level <= self.rollbar_level:
                    try:
                        self.notifier.report_message(message, level)
                        if exc_info[0]:
                            self.notifier.report_exc_info(exc_info)
                    except:
                        self.logger.error("Could not send the message to rollbar.")
            except:
                self.logger.error("Wrong rollbar level specified.")

class replica_engine(object):
    """
        This class is wraps the the mysql and postgresql engines in order to perform the various activities required for the replica.
        The constructor inits the global configuration class  and setup the mysql and postgresql engines as class objects.
        The class sets the logging using the configuration parameter.

    """
    def __init__(self, args):
        """
            Class constructor.
        """
        if os.geteuid()==0:
            print ("pg_chameleon cannot be run as root")
            sys.exit(10)


        self.catalog_version = '2.0.9'
        self.upgradable_version = '1.7'
        self.lst_yes= ['yes',  'Yes', 'y', 'Y']
        python_lib=os.path.dirname(os.path.realpath(__file__))

        cham_dir = "%s/.pg_chameleon" % os.path.expanduser('~')


        local_conf = "%s/configuration/" % cham_dir
        self.global_conf_example = '%s/../configuration/config-example.yml' % python_lib
        self.local_conf_example = '%s/config-example.yml' % local_conf

        local_logs = "%s/logs/" % cham_dir
        local_pid = "%s/pid/" % cham_dir

        self.conf_dirs=[
            cham_dir,
            local_conf,
            local_logs,
            local_pid,

        ]
        self.args = args
        self.source = self.args.source
        if self.args.command == 'set_configuration_files':
            self.set_configuration_files()
            sys.exit()

        self.__set_conf_permissions(cham_dir)

        self.load_config()
        log_list = self.__init_logger("global")
        self.logger = log_list[0]
        self.logger_fds = log_list[1]

        #notifier configuration
        self.notifier = rollbar_notifier(self.config["rollbar_key"],self.config["rollbar_env"] , self.args.rollbar_level ,  self.logger )

        #pg_engine instance initialisation
        self.pg_engine = pg_engine()
        self.pg_engine.dest_conn = self.config["pg_conn"]
        self.pg_engine.logger = self.logger
        self.pg_engine.source = self.args.source
        self.pg_engine.full = self.args.full
        self.pg_engine.type_override = self.config["type_override"]
        self.pg_engine.sources = self.config["sources"]
        self.pg_engine.notifier = self.notifier
        self.pg_engine.fillfactor = self.config["fillfactor"]
        


        #mysql_source instance initialisation
        self.mysql_source = mysql_source()
        self.mysql_source.source = self.args.source
        self.mysql_source.tables = self.args.tables
        self.mysql_source.schema = self.args.schema.strip()
        self.mysql_source.pg_engine = self.pg_engine
        self.mysql_source.logger = self.logger
        self.mysql_source.sources = self.config["sources"]
        self.mysql_source.type_override = self.config["type_override"]
        self.mysql_source.notifier = self.notifier


        #pgsql_source instance initialisation
        self.pgsql_source = pgsql_source()
        self.pgsql_source.source = self.args.source
        self.pgsql_source.tables = self.args.tables
        self.pgsql_source.schema = self.args.schema.strip()
        self.pgsql_source.pg_engine = self.pg_engine
        self.pgsql_source.logger = self.logger
        self.pgsql_source.sources = self.config["sources"]
        self.pgsql_source.type_override = self.config["type_override"]
        self.pgsql_source.notifier = self.notifier

        catalog_version = self.pg_engine.get_catalog_version()
        #safety checks
        if self.args.command == 'upgrade_replica_schema':
            self.pg_engine.sources = self.config["sources"]
            print("WARNING, entering upgrade mode. Disabling the catalogue version's check. Expected version %s, installed version %s" % (self.catalog_version, catalog_version))
        elif self.args.command == 'enable_replica' and self.catalog_version != catalog_version:
            print("WARNING, catalogue mismatch. Expected version %s, installed version %s" % (self.catalog_version, catalog_version))
        else:
            if  catalog_version:
                if self.catalog_version != catalog_version:
                    print("FATAL, replica catalogue version mismatch. Expected %s, got %s" % (self.catalog_version, catalog_version))
                    sys.exit()

        if self.args.source != '*' and self.args.command != 'add_source':
            self.pg_engine.connect_db()
            source_count = self.pg_engine.check_source()
            self.pg_engine.disconnect_db()
            if source_count == 0:
                print("FATAL, The source %s is not registered. Please add it with the command add_source" % (self.args.source))
                sys.exit()




    def terminate_replica(self, signal, frame):
        """
            Stops gracefully the replica.
        """
        self.logger.info("Caught stop replica signal terminating daemons and ending the replica process.")
        self.read_daemon.terminate()
        self.replay_daemon.terminate()
        self.pg_engine.connect_db()
        self.pg_engine.set_source_status("stopped")
        sys.exit(0)

    def set_configuration_files(self):
        """
            The method loops the list self.conf_dirs creating them only if they are missing.

            The method checks the freshness of the config-example.yaml and connection-example.yml
            copies the new version from the python library determined in the class constructor with get_python_lib().

            If the configuration file is missing the method copies the file with a different message.

        """

        for confdir in self.conf_dirs:
            if not os.path.isdir(confdir):
                print ("creating directory %s" % confdir)
                os.mkdir(confdir)


        if os.path.isfile(self.local_conf_example):
            if os.path.getctime(self.global_conf_example)>os.path.getctime(self.local_conf_example):
                print ("updating configuration example with %s" % self.local_conf_example)
                copy(self.global_conf_example, self.local_conf_example)
        else:
            print ("copying configuration  example in %s" % self.local_conf_example)
            copy(self.global_conf_example, self.local_conf_example)

    def load_config(self):
        """
            The method loads the configuration from the file specified in the args.config parameter.
        """
        local_confdir = "%s/.pg_chameleon/configuration/" % os.path.expanduser('~')
        self.config_file = '%s/%s.yml'%(local_confdir, self.args.config)

        if not os.path.isfile(self.config_file):
            print("**FATAL - configuration file missing. Please ensure the file %s is present." % (self.config_file))
            sys.exit()

        config_file = open(self.config_file, 'r')
        pyyml=str(get_distribution('PyYAML')).split(' ')[1]
        if pyyml<='3.13':
            self.config = yaml.load(config_file.read())
        else:
            self.config = yaml.load(config_file.read(), Loader=yaml.FullLoader)
        config_file.close()

        #managing default values for optional keys
        if "fillfactor" not in self.config:
            self.config["fillfactor"] = None
            
        if "type_override" not in self.config:
            self.config["type_override"] = None
            


    def show_sources(self):
        """
            The method shows the sources available in the configuration file.
        """
        for item in self.config["sources"]:
            print("\n")
            print (tabulate([], headers=["Source %s" % item]))
            tab_headers = ['Parameter', 'Value']
            tab_body = []
            source = self.config["sources"][item]
            config_list = [param for param in source if param not in ['db_conn']]
            connection_list = [param for param  in source["db_conn"] if param not in ['password']]
            for parameter in config_list:
                tab_row = [parameter, source[parameter]]
                tab_body.append(tab_row)
            for param in connection_list:
                tab_row = [param, source["db_conn"][param]]
                tab_body.append(tab_row)

            print(tabulate(tab_body, headers=tab_headers))

    def show_config(self):
        """
            The method loads the current configuration and displays the status in tabular output
        """
        config_list = [item for item in self.config if item not in ['pg_conn', 'sources', 'type_override']]
        connection_list = [item for item in self.config["pg_conn"] if item not in ['password']]
        type_override = pprint.pformat(self.config['type_override'], width = 20)
        tab_body = []
        tab_headers = ['Parameter', 'Value']
        for item in config_list:
            tab_row = [item, self.config[item]]
            tab_body.append(tab_row)
        for item in connection_list:
            tab_row = [item, self.config["pg_conn"][item]]
            tab_body.append(tab_row)
        tab_row = ['type_override', type_override]
        tab_body.append(tab_row)
        print(tabulate(tab_body, headers=tab_headers))
        self.show_sources()

    def create_replica_schema(self):
        """
            The method creates the replica schema in the destination database.
        """
        self.logger.info("Trying to create replica schema")
        self.pg_engine.create_replica_schema()

    def drop_replica_schema(self):
        """
            The method removes the replica schema from the destination database.
        """
        self.logger.info("Dropping the replica schema")
        self.pg_engine.drop_replica_schema()

    def add_source(self):
        """
            The method adds a new replication source. A pre existence check is performed
        """
        if self.args.source == "*":
            print("You must specify a source name with the argument --source")
        else:
            self.logger.info("Trying to add a new source")
            self.pg_engine.add_source()

    def drop_source(self):
        """
            The method removes a replication source from the catalogue.
        """
        if self.args.source == "*":
            print("You must specify a source name with the argument --source")
        else:
            drp_msg = 'Dropping the source %s will remove drop any replica reference.\n Are you sure? YES/No\n'  % self.args.source
            drop_src = input(drp_msg)
            if drop_src == 'YES':
                self.logger.info("Trying to remove the source")
                self.pg_engine.drop_source()
            elif drop_src in  self.lst_yes:
                print('Please type YES all uppercase to confirm')


    def enable_replica(self):
        """
            The method  resets the source status to stopped and disables any leftover maintenance mode
        """
        self.pg_engine.connect_db()
        self.pg_engine.set_source_status("stopped")
        self.pg_engine.end_maintenance()

    def copy_schema(self):
        """
            The method calls init_replica adding a flag for skipping the data copy.
            Useful if we want to test for schema issues or to populate the schema preventively.
            
        """
        self.mysql_source.copy_table_data=False
        self.init_replica()
        self.pg_engine.fk_metadata = self.mysql_source.get_foreign_keys_metadata()
        self.pg_engine.create_foreign_keys()

    def init_replica(self):
        """
            The method  initialise a replica for a given source and configuration.
            It is compulsory to specify a source name when running this method.
            The method checks the source type and calls the corresponding initialisation's method.

        """
        if self.args.source == "*":
            print("You must specify a source name with the argument --source")
        elif self.args.tables != "*":
            print("You cannot specify a table name when running init_replica.")
        else:
            try:
                source_type = self.config["sources"][self.args.source]["type"]
            except KeyError:
                print("The source %s doesn't exists." % (self.args.source))
                sys.exit()
            self.__stop_replica()
            if source_type  == "mysql":
                self.__init_mysql_replica()
            elif source_type  == "pgsql":
                self.__init_pgsql_replica()

    def __init_mysql_replica(self):
        """
            The method  initialise a replica for a given mysql source within the specified configuration.
            The method is called by the public method init_replica.
        """
        if self.args.debug:
            self.mysql_source.init_replica()
        else:
            if self.config["log_dest"]  == 'stdout':
                foreground = True
            else:
                foreground = False
                print("Process for source %s started." % (self.args.source))
            keep_fds = [self.logger_fds]
            init_pid = os.path.expanduser('%s/%s.pid' % (self.config["pid_dir"],self.args.source))
            self.logger.info("Initialising the replica for source %s" % self.args.source)
            init_daemon = Daemonize(app="init_replica", pid=init_pid, action=self.mysql_source.init_replica, foreground=foreground , keep_fds=keep_fds)
            init_daemon.start()

    def __init_pgsql_replica(self):
        """
            The method  initialise a replica for a given postgresql source within the specified configuration.
            The method is called by the public method init_replica.
        """

        if self.args.debug:
            self.pgsql_source.init_replica()
        else:
            if self.config["log_dest"]  == 'stdout':
                foreground = True
            else:
                foreground = False
                print("Process for source %s started." % (self.args.source))
            keep_fds = [self.logger_fds]
            init_pid = os.path.expanduser('%s/%s.pid' % (self.config["pid_dir"],self.args.source))
            self.logger.info("Initialising the replica for source %s" % self.args.source)
            init_daemon = Daemonize(app="init_replica", pid=init_pid, action=self.pgsql_source.init_replica, foreground=foreground , keep_fds=keep_fds)
            init_daemon.start()


    def refresh_schema(self):
        """
            The method  reload the data from a source and only for a specified schema.
            Is compulsory to specify a source name and an origin's schema name.
            The schema mappings are honoured by the procedure automatically.
        """
        if self.args.source == "*":
            print("You must specify a source name using the argument --source")
        elif self.args.schema == "*":
            print("You must specify an origin's schema name using the argument --schema")
        else:
            self.__stop_replica()
            if self.args.debug:
                self.mysql_source.refresh_schema()
            else:
                if self.config["log_dest"]  == 'stdout':
                    foreground = True
                else:
                    foreground = False
                    print("Sync tables process for source %s started." % (self.args.source))
                keep_fds = [self.logger_fds]
                init_pid = os.path.expanduser('%s/%s.pid' % (self.config["pid_dir"],self.args.source))
                self.logger.info("The tables %s within source %s will be synced." % (self.args.tables, self.args.source))
                sync_daemon = Daemonize(app="sync_tables", pid=init_pid, action=self.mysql_source.refresh_schema, foreground=foreground , keep_fds=keep_fds)
                sync_daemon .start()


    def sync_tables(self):
        """
            The method  reload the data from a source only for specified tables.
            Is compulsory to specify a source name and at least one table name when running this method.
            Multiple tables are allowed if comma separated.
        """
        if self.args.source == "*":
            print("You must specify a source name using the argument --source")
        elif self.args.tables == "*":
            print("You must specify one or more tables, in the form schema.table, separated by comma using the argument --tables")
        else:
            self.__stop_replica()
            if self.args.debug:
                self.mysql_source.sync_tables()
            else:
                if self.config["log_dest"]  == 'stdout':
                    foreground = True
                else:
                    foreground = False
                    print("Sync tables process for source %s started." % (self.args.source))
                keep_fds = [self.logger_fds]
                init_pid = os.path.expanduser('%s/%s.pid' % (self.config["pid_dir"],self.args.source))
                self.logger.info("The tables %s within source %s will be synced." % (self.args.tables, self.args.source))
                sync_daemon = Daemonize(app="sync_tables", pid=init_pid, action=self.mysql_source.sync_tables, foreground=foreground , keep_fds=keep_fds)
                sync_daemon .start()

    def __stop_all_active_sources(self):
        """
            The method stops all the active sources within the target PostgreSQL database.
        """
        active_source = self.pg_engine.get_active_sources()
        for source in active_source:
            self.source = source[0]
            self.__stop_replica()

    def upgrade_replica_schema(self):
        """
            The method upgrades an existing replica catalogue to the newer version.
            If the catalogue is from the previous version
        """
        catalog_version = self.pg_engine.get_catalog_version()
        if catalog_version == self.catalog_version:
            print("The replica catalogue is already up to date.")
            sys.exit()
        else:
            if catalog_version == self.upgradable_version:
                upg_msg = 'Upgrading the catalogue %s to the version %s.\n Are you sure? YES/No\n'  %  (catalog_version, self.catalog_version)
                upg_cat = input(upg_msg)
                if upg_cat == 'YES':
                    self.logger.info("Performing the upgrade")
                    self.pg_engine.upgrade_catalogue_v1()
                elif upg_cat in  self.lst_yes:
                    print('Please type YES all uppercase to confirm')
            elif catalog_version.split('.')[0] == '2' and catalog_version.split('.')[1] == '0':
                print('Stopping all the active sources.')
                self.__stop_all_active_sources()
                print('Upgrading the replica catalogue. ')
                self.pg_engine.upgrade_catalogue_v20()
            else:
                print('Wrong starting version. Expected %s, got %s' % (catalog_version, self.upgradable_version))
                sys.exit()


    def update_schema_mappings(self):
        """
            The method updates the schema mappings for the given source.
            The schema mappings is a configuration parameter but is stored in the replica
            catalogue when the source is added. If any change is made on the configuration file this method
            should be called to update the system catalogue as well. The pg_engine method checks for any conflict before running
            the update on the tables t_sources and t_replica_tables.
            Is compulsory to specify a source name when running this method.
        """
        if self.args.source == "*":
            print("You must specify a source name with the argument --source")
        else:
            self.__stop_replica()
            self.pg_engine.update_schema_mappings()


    def read_replica(self, queue, log_read):
        """
            The method reads the replica stream for the given source and stores the row images
            in the target postgresql database.
        """
        if "keep_existing_schema" in self.config["sources"][self.args.source]:
            keep_existing_schema = self.config["sources"][self.args.source]["keep_existing_schema"]
        else:
            keep_existing_schema = False
        self.mysql_source.keep_existing_schema = keep_existing_schema
        self.mysql_source.logger  = log_read[0]
        self.pg_engine.logger  = log_read[0]
        while True:
            try:
                self.mysql_source.read_replica()
                time.sleep(self.sleep_loop)
            except Exception:
                queue.put(traceback.format_exc())
                break

    def replay_replica(self, queue, log_replay):
        """
            The method replays the row images stored in the target postgresql database.
        """
        self.pg_engine.logger  = log_replay[0]
        tables_error  = []
        self.pg_engine.connect_db()
        self.pg_engine.set_source_id()
        while True:
            try:
                tables_error = self.pg_engine.replay_replica()
                if len(tables_error) > 0:
                    table_list = [item for sublist in tables_error for item in sublist]
                    tables_removed = "\n".join(table_list)
                    notifier_message = "There was an error during the replay of data. %s. The affected tables are no longer replicated." % (tables_removed)
                    self.logger.error(notifier_message)
                    self.notifier.send_message(notifier_message, 'error')

            except Exception:
                queue.put(traceback.format_exc())
                break
            time.sleep(self.sleep_loop)

    def __run_replica(self):
        """
            This method is the method which manages the two separate processes using the multiprocess library.
            It can be daemonised or run in foreground according with the --debug configuration or the log
            destination.
        """
        if "auto_maintenance" not in  self.config["sources"][self.args.source]:
            auto_maintenance = "disabled"
        else:
            auto_maintenance = self.config["sources"][self.args.source]["auto_maintenance"]

        if "gtid_enable" not in  self.config["sources"][self.args.source]:
            gtid_enable = False
        else:
            gtid_enable =  self.config["sources"][self.args.source]["gtid_enable"]

        self.mysql_source.gtid_enable = gtid_enable


        log_read = self.__init_logger("read")
        log_replay = self.__init_logger("replay")

        signal.signal(signal.SIGINT, self.terminate_replica)
        queue = mp.Queue()
        self.sleep_loop = self.config["sources"][self.args.source]["sleep_loop"]
        if self.args.debug:
            check_timeout = self.sleep_loop
        else:
            check_timeout = self.sleep_loop*10
        self.logger.info("Starting the replica daemons for source %s " % (self.args.source))
        self.read_daemon = mp.Process(target=self.read_replica, name='read_replica', daemon=True, args=(queue, log_read,))
        self.replay_daemon = mp.Process(target=self.replay_replica, name='replay_replica', daemon=True, args=(queue, log_replay,))
        self.read_daemon.start()
        self.replay_daemon.start()
        while True:
            read_alive = self.read_daemon.is_alive()
            replay_alive = self.replay_daemon.is_alive()
            if  read_alive and replay_alive:
                self.logger.debug("Replica process for source %s is running" % (self.args.source))
                self.pg_engine.cleanup_replayed_batches()
            else:
                try:
                    stack_trace = queue.get_nowait()
                except:
                    stack_trace = 'Stack trace not available'
                self.logger.error("Read process alive: %s - Replay process alive: %s" % (read_alive, replay_alive, ))
                self.logger.error("Stack trace: %s" % (stack_trace, ))
                if read_alive:
                    self.read_daemon.terminate()
                    self.logger.error("Replay daemon crashed. Terminating the read daemon.")
                if replay_alive:
                    self.replay_daemon.terminate()
                    self.logger.error("Read daemon crashed. Terminating the replay daemon.")
                if self.args.debug:
                    replica_status = "stopped"
                else:
                    replica_status = "error"
                try:
                    self.pg_engine.connect_db()
                    self.pg_engine.set_source_status(replica_status)
                except:
                    pass
                notifier_message = "The replica process crashed.\n Source: %s\n Stack trace: %s " %(self.args.source, stack_trace)
                self.notifier.send_message(notifier_message, 'critical')

                break
            time.sleep(check_timeout)
            if auto_maintenance != "disabled":
                self.pg_engine.auto_maintenance = auto_maintenance
                self.pg_engine.connect_db()
                run_maintenance = self.pg_engine.check_auto_maintenance()
                self.pg_engine.disconnect_db()
                if run_maintenance:
                    self.pg_engine.run_maintenance()



        self.logger.info("Replica process for source %s ended" % (self.args.source))

    def start_replica(self):
        """
            The method starts a new replica process.
            Is compulsory to specify a source name when running this method.
        """

        replica_pid = os.path.expanduser('%s/%s.pid' % (self.config["pid_dir"],self.args.source))

        if self.args.source == "*":
            print("You must specify a source name using the argument --source")
        else:
            self.pg_engine.connect_db()
            self.logger.info("Checking if the replica for source %s is stopped " % (self.args.source))
            replica_status = self.pg_engine.get_replica_status()
            if replica_status in ['syncing', 'running', 'initialising']:
                print("The replica process is already started or is syncing. Aborting the command.")
            elif replica_status == 'error':
                print("The replica process is in error state.")
                print("You may need to check the replica status first. To enable it run the following command.")
                print("chameleon.py enable_replica --config %s --source %s " % (self.args.config, self.args.source))

            else:
                self.logger.info("Cleaning not processed batches for source %s" % (self.args.source))
                self.pg_engine.clean_not_processed_batches()
                self.pg_engine.disconnect_db()
                if self.args.debug:
                    self.__run_replica()
                else:
                    if self.config["log_dest"]  == 'stdout':
                        foreground = True
                    else:
                        foreground = False
                        print("Starting the replica process for source %s" % (self.args.source))

                    keep_fds = [self.logger_fds]
                    app_name = "%s_replica" % self.args.source
                    replica_daemon = Daemonize(app=app_name, pid=replica_pid, action=self.__run_replica, foreground=foreground , keep_fds=keep_fds)
                    try:
                        replica_daemon.start()
                    except:
                        print("The replica process is already started. Aborting the command.")


    def __stop_replica(self):
        """
            The method reads the pid of the replica process for the given self.source and sends a SIGINT which
            tells the replica process to manage a graceful exit.
        """
        replica_pid = os.path.expanduser('%s/%s.pid' % (self.config["pid_dir"],self.source))
        if os.path.isfile(replica_pid):
            try:
                file_pid=open(replica_pid,'r')
                pid=file_pid.read()
                file_pid.close()
                os.kill(int(pid),2)
                print("Requesting the replica for source %s to stop" % (self.source))
                while True:
                    try:
                        os.kill(int(pid),0)
                    except:
                        break
                print("The replica process is stopped")
            except:
                print("An error occurred when trying to signal the replica process")

    def __set_conf_permissions(self,  cham_dir):
        """
            The method sets the permissions of the configuration directory to 700

            :param cham_dir: the chameleon configuration directory to fix
        """
        if os.path.isdir(cham_dir):
            os.chmod(cham_dir, 0o700)


    def stop_replica(self):
        """
            The method calls the private method __stop_replica to stop the replica process.
        """
        self.__stop_replica()

    def stop_all_replicas(self):
        """
            The method  stops all the active replicas within the target database
        """
        self.__stop_all_active_sources()

    def show_errors(self):
        """
            displays the error log entries if any.
            If the source the error log is filtered for this source only.
        """
        log_id = self.args.logid
        self.pg_engine.source = self.args.source
        log_error_data = self.pg_engine.get_log_data(log_id)
        if log_error_data:
            if log_id != "*":
                tab_body = []
                log_line = log_error_data[0]
                tab_body.append(['Log id', log_line[0]])
                tab_body.append(['Source name', log_line[1]])
                tab_body.append(['ID Batch', log_line[2]])
                tab_body.append(['Table', log_line[3]])
                tab_body.append(['Schema', log_line[4]])
                tab_body.append(['Error timestamp', log_line[5]])
                tab_body.append(['SQL executed', log_line[6]])
                tab_body.append(['Error message', log_line[7]])
                print(tabulate(tab_body, tablefmt="simple"))
            else:
                tab_headers = ['Log id',  'Source name', 'ID Batch',  'Table', 'Schema' ,  'Error timestamp']
                tab_body = []
                for log_line in log_error_data:
                    log_id = log_line[0]
                    id_batch = log_line[1]
                    source_name = log_line[2]
                    table_name = log_line[3]
                    schema_name = log_line[4]
                    error_timestamp = log_line[5]
                    tab_row = [log_id, id_batch,source_name, table_name,   schema_name, error_timestamp]
                    tab_body.append(tab_row)
                print(tabulate(tab_body, headers=tab_headers, tablefmt="simple"))
        else:
            print('There are no errors in the log')

    def show_status(self):
        """
            list the replica status from the replica catalogue.
            If the source is specified gives some extra details on the source status.
        """
        self.pg_engine.auto_maintenance = "disabled"
        if self.args.source != "*":
            if "auto_maintenance" in  self.config["sources"][self.args.source]:
                self.pg_engine.auto_maintenance = self.config["sources"][self.args.source]["auto_maintenance"]

        self.pg_engine.source = self.args.source
        configuration_data = self.pg_engine.get_status()
        configuration_status = configuration_data[0]
        schema_mappings = configuration_data[1]
        table_status = configuration_data[2]
        replica_counters = configuration_data[3]
        tab_headers = ['Source id',  'Source name', 'Type',  'Status', 'Consistent' ,  'Read lag',  'Last read',  'Replay lag' , 'Last replay']
        tab_body = []
        for status in configuration_status:
            source_id = status[0]
            source_name = status[1]
            source_status = status[2]
            read_lag = status[3]
            last_read = status[4]
            replay_lag = status[5]
            last_replay = status[6]
            consistent = status[7]
            source_type = status[8]
            last_maintenance = status[9]
            next_maintenance = status[10]
            tab_row = [source_id, source_name, source_type,   source_status, consistent,  read_lag, last_read,  replay_lag, last_replay]
            tab_body.append(tab_row)
        print(tabulate(tab_body, headers=tab_headers, tablefmt="simple"))
        if schema_mappings:
            print('\n== Schema mappings ==')
            tab_headers = ['Origin schema',  'Destination schema']
            tab_body = []
            for mapping in schema_mappings:
                origin_schema =  mapping[0]
                destination_schema=  mapping[1]
                tab_row = [origin_schema, destination_schema]
                tab_body.append(tab_row)
            print(tabulate(tab_body, headers=tab_headers, tablefmt="simple"))
        if table_status:
            print('\n== Replica status ==')
            #tab_headers = ['',  '',  '']
            tab_body = []
            tables_no_replica = table_status[0]
            tab_row = ['Tables not replicated', tables_no_replica[1]]
            tab_body.append(tab_row)
            tables_with_replica = table_status[1]
            tab_row = ['Tables replicated', tables_with_replica[1]]
            tab_body.append(tab_row)
            tables_all= table_status[2]
            tab_row = ['All tables', tables_all[1]]
            tab_body.append(tab_row)
            tab_row = ['Last maintenance', last_maintenance]
            tab_body.append(tab_row)
            tab_row = ['Next maintenance', next_maintenance]
            tab_body.append(tab_row)
            if replica_counters:
                tab_row = ['Replayed rows', replica_counters[0]]
                tab_body.append(tab_row)
                tab_row = ['Replayed DDL', replica_counters[2]]
                tab_body.append(tab_row)
                tab_row = ['Skipped rows', replica_counters[1]]
                tab_body.append(tab_row)
            print(tabulate(tab_body, tablefmt="simple"))
            if tables_no_replica[2]:
                print('\n== Tables with replica disabled ==')
                print("\n".join(tables_no_replica[2]))

    def detach_replica(self):
        """
            The method terminates the replica process. The source is removed from the table t_sources with all the associated data.
            The schema sequences in are reset to the max values in the corresponding tables, leaving
            the postgresql database as a standalone snapshot.
            The method creates the foreign keys existing in MySQL as well.
            Is compulsory to specify a source name when running this method.
        """
        if self.args.source == "*":
            print("You must specify a source name with the argument --source")
        elif self.args.tables != "*":
            print("You cannot specify a table name when running detach_replica.")
        else:
            drp_msg = 'Detaching the replica will remove any reference for the source %s.\n Are you sure? YES/No\n'  % self.args.source
            drop_src = input(drp_msg)
            if drop_src == 'YES':
                if "keep_existing_schema" in self.config["sources"][self.args.source]:
                    keep_existing_schema = self.config["sources"][self.args.source]["keep_existing_schema"]
                else:
                    keep_existing_schema = False
                self.pg_engine.keep_existing_schema = keep_existing_schema
                if not keep_existing_schema:
                    self.pg_engine.fk_metadata = self.mysql_source.get_foreign_keys_metadata()
                self.__stop_replica()
                self.pg_engine.detach_replica()
            elif drop_src in self.lst_yes:
                print('Please type YES all uppercase to confirm')

    def run_maintenance(self):
        """
            The method runs a maintenance process on the target postgresql database specified in the given source.
        """
        maintenance_pid = os.path.expanduser('%s/%s_maintenance.pid' % (self.config["pid_dir"],self.args.source))
        if self.args.source == "*":
            print("You must specify a source name with the argument --source")
        else:
            if self.args.debug:
                self.pg_engine.run_maintenance()
            else:
                if self.config["log_dest"]  == 'stdout':
                    foreground = True
                else:
                    self.logger.info("Starting the maintenance on the source %s" % (self.args.source, ))
                    foreground = False
                    print("Starting the maintenance process for source %s" % (self.args.source))
                    keep_fds = [self.logger_fds]

                    app_name = "%s_maintenance" % self.args.source
                    maintenance_daemon = Daemonize(app=app_name, pid=maintenance_pid, action=self.pg_engine.run_maintenance, foreground=foreground , keep_fds=keep_fds)
                    try:
                        maintenance_daemon.start()
                    except:
                        print("The  maintenance process is already started. Aborting the command.")

    def __init_logger(self, logger_name):
        """
        The method initialise a new logger object using the configuration parameters.
        The formatter is different if the debug option is enabler or not.
        The method returns a new logger object and sets the logger's file descriptor in the class variable
        logger_fds, used when the process is demonised.

        :param logger_name: the name of the logger used to build the file name and get the correct logger
        :return: list with logger and file descriptor
        :rtype: list

        """
        log_dir = self.config["log_dir"]
        log_level = self.config["log_level"]
        log_dest = self.config["log_dest"]
        log_days_keep = self.config["log_days_keep"]
        config_name = self.args.config
        source_name = self.args.source
        debug_mode = self.args.debug
        if source_name == '*':
            log_name = "%s_general" % (config_name)
        elif  logger_name == "global":
            log_name = "%s_%s" % (config_name, source_name)
        else:
            log_name = "%s_%s_%s" % (config_name, source_name, logger_name)

        log_file = os.path.expanduser('%s/%s.log' % (log_dir,log_name))
        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.DEBUG)
        logger.propagate = False
        if debug_mode:
            str_format = "%(asctime)s %(processName)s %(levelname)s %(filename)s (%(lineno)s): %(message)s"
        else:
            str_format = "%(asctime)s %(processName)s %(levelname)s: %(message)s"
        formatter = logging.Formatter(str_format, "%Y-%m-%d %H:%M:%S")

        if log_dest=='stdout' or debug_mode:
            fh=logging.StreamHandler(sys.stdout)

        elif log_dest=='file':
            fh = TimedRotatingFileHandler(log_file, when="d",interval=1,backupCount=log_days_keep)

        if log_level=='debug' or debug_mode:
            fh.setLevel(logging.DEBUG)
        elif log_level=='info':
            fh.setLevel(logging.INFO)

        fh.setFormatter(formatter)
        logger.addHandler(fh)
        logger_fds = fh.stream.fileno()
        return [logger, logger_fds]
